# -*- coding: utf-8 -*-

# Copyright © 2017-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core import ChainArray
from .base import *
from .gradient import *
from .standard_d import interruptible_minimize
from warnings import warn
from math import pi, log
import numpy as np
import pandas as pd
#from scipy.optimize import minimize
from collections import OrderedDict


setup = {'name': ('standard.df', 'smooth.df'),
    'provides': 'df',
    'arguments': OrderedDict((
        ('localization_error',  ('-e', dict(type=float, help='localization precision (see also sigma; default is 0.03)'))),
        ('diffusivity_prior',   ('-d', dict(type=float, help='prior on the diffusivity'))),
        #('potential_prior',     ('-v', dict(type=float, help='prior on the potential (use force_prior instead)'))),
        ('force_prior',         ('-f', dict(type=float, help='prior on the amplitude of the force'))),
        ('jeffreys_prior',      ('-j', dict(action='store_true', help="Jeffreys' prior"))),
        ('min_diffusivity',     dict(type=float, help='minimum diffusivity value allowed')),
        ('max_iter',        dict(type=int, help='maximum number of iterations')),
        ('rgrad',       dict(help="alternative gradient for the regularization; can be 'delta'/'delta0' or 'delta1'")),
        ('tol',             dict(type=float, help='tolerance for scipy minimizer')))),
    'cell_sampling': 'group'}
setup_with_grad_arguments(setup)


def smooth_df_neg_posterior(x, df, cells, sigma2, diffusivity_prior,
        force_prior, jeffreys_prior, dt_mean, min_diffusivity,
        index, reverse_index, grad_kwargs):
    # extract `D` and `F`
    df.update(x)
    D, F = df['D'], df['F']
    #
    if min_diffusivity is not None:
        observed_min = np.min(D)
        if observed_min < min_diffusivity and not np.isclose(observed_min, min_diffusivity):
            warn(DiffusivityWarning(observed_min, min_diffusivity))
    noise_dt = sigma2
    # for all cell
    result = 0.
    for j, i in enumerate(index):
        cell = cells[i]
        n = len(cell) # number of translocations
        # various posterior terms
        D_dt = D[j] * cell.dt
        denominator = 4. * (D_dt + noise_dt) # 4*(D+Dnoise)*dt
        dr_minus_drift_dt = cell.dr - np.outer(D_dt, F[j])
        # non-directional squared displacement
        ndsd = np.sum(dr_minus_drift_dt * dr_minus_drift_dt, axis=1)
        result += n * log(pi) + np.sum(np.log(denominator)) + np.sum(ndsd / denominator)
        # priors
        if diffusivity_prior:
            gradD = cells.grad(i, D, reverse_index, **grad_kwargs) # spatial gradient of the local diffusivity
            if gradD is not None:
                # `grad_sum` memoizes and can be called several times at no extra cost
                result += diffusivity_prior * cells.grad_sum(i, gradD * gradD)
        if force_prior:
            result += force_prior * cells.grad_sum(i, F * F)
    if jeffreys_prior:
        result += 2. * np.sum(np.log(D * dt_mean + sigma2) - np.log(D))
    return result


def df_neg_posterior1(x, df, cells, sigma2, diffusivity_prior,
        force_prior, jeffreys_prior, dt_mean, min_diffusivity,
        index, reverse_index, grad_kwargs):
    # extract `D` and `F`
    df.update(x)
    D, F = df['D'], df['F']
    #
    if min_diffusivity is not None:
        observed_min = np.min(D)
        if observed_min < min_diffusivity and not np.isclose(observed_min, min_diffusivity):
            warn(DiffusivityWarning(observed_min, min_diffusivity))
    noise_dt = sigma2
    # for all cell
    result = 0.
    for j, i in enumerate(index):
        cell = cells[i]
        n = len(cell) # number of translocations
        # various posterior terms
        D_dt = D[j] * cell.dt
        denominator = 4. * (D_dt + noise_dt) # 4*(D+Dnoise)*dt
        dr_minus_drift_dt = cell.dr - np.outer(D_dt, F[j])
        # non-directional squared displacement
        ndsd = np.sum(dr_minus_drift_dt * dr_minus_drift_dt, axis=1)
        result += n * log(pi) + np.sum(np.log(denominator)) + np.sum(ndsd / denominator)
        # priors
        if diffusivity_prior:
            deltaD = cells.local_variation(i, D, reverse_index, **grad_kwargs) # spatial gradient of the local diffusivity
            if deltaD is not None:
                # `grad_sum` memoizes and can be called several times at no extra cost
                result += diffusivity_prior * cells.grad_sum(i, deltaD * deltaD)
        if force_prior:
            result += force_prior * cells.grad_sum(i, F * F)
    if jeffreys_prior:
        result += 2. * np.sum(np.log(D * dt_mean + sigma2) - np.log(D))
    return result


def infer_smooth_DF(cells, diffusivity_prior=None, force_prior=None, potential_prior=None,
        jeffreys_prior=False, min_diffusivity=None, max_iter=None, epsilon=None, rgrad=None,
        verbose=False, **kwargs):
    """
    Argument `potential_prior` is an alias for `force_prior` which penalizes the large force amplitudes.
    """

    # initial values
    localization_error = cells.get_localization_error(kwargs, 0.03, True)
    index, reverse_index, n, dt_mean, D_initial, min_diffusivity, D_bounds, _ = \
        smooth_infer_init(cells, min_diffusivity=min_diffusivity, jeffreys_prior=jeffreys_prior,
        sigma2=localization_error)
    F_initial = np.zeros((len(index), cells.dim), dtype=D_initial.dtype)
    df = ChainArray('D', D_initial, 'F', F_initial)

    # gradient options
    grad_kwargs = get_grad_kwargs(epsilon=epsilon, **kwargs)

    # parametrize the optimization algorithm
    default_lBFGSb_options = dict(maxiter=1e3, maxfun=1e10, ftol=1e-6)
    # in L-BFGS-B the number of iterations is usually very low (~10-100) while the number of
    # function evaluations is much higher (~1e4-1e5);
    # with maxfun defined, an iteration can stop anytime and the optimization may terminate
    # with an error message
    if min_diffusivity is None:
        options = {}
    else:
        F_bounds = [(None, None)] * F_initial.size # no bounds
        kwargs['bounds'] = D_bounds + F_bounds
        options = dict(default_lBFGSb_options)
    options.update(kwargs.pop('options', {}))
    if max_iter:
        options['maxiter'] = max_iter
    if verbose:
        options['disp'] = verbose
    if options:
        kwargs['options'] = options

    # posterior function
    if rgrad in ('delta','delta0','delta1'):
        fun = df_neg_posterior1
    else:
        if rgrad not in (None, 'grad', 'grad1', 'gradn'):
            warn('unsupported rgrad: {}'.format(rgrad), RuntimeWarning)
        fun = smooth_df_neg_posterior

    if force_prior is None:
        if potential_prior is not None:
            raise ValueError('potential prior: {}'.format(potential_prior))
            warn('please use `force_prior` instead of `potential_prior`', PendingDeprecationWarning)
        force_prior = potential_prior

    # run the optimization
    #cell.cache = None # no cache needed
    args = (df, cells, localization_error, diffusivity_prior, force_prior, jeffreys_prior, dt_mean, min_diffusivity, index, reverse_index, grad_kwargs)
    result = interruptible_minimize(fun, df.combined, args=args, **kwargs)
    if not (result.success or verbose):
        warn('{}'.format(result.message), OptimizationWarning)

    # collect the result
    df.update(result.x)
    D, F = df['D'], df['F']
    DF = pd.DataFrame(np.concatenate((D[:,np.newaxis], F), axis=1), index=index, \
        columns=[ 'diffusivity' ] + \
            [ 'force ' + col for col in cells.space_cols ])

    return DF




# -*- coding: utf-8 -*-

# Copyright © 2018-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .base import *
from .gradient import *
from .dv import DV
from .optimization import *
from tramway.core import parallel
from math import pi, log
import numpy as np
import pandas as pd
import scipy.sparse as sparse
from collections import OrderedDict, deque
import time
from scipy.stats import trim_mean
import logging
import os


setup = {'name': ('stochastic.dv', 'stochastic.dv1'),
    'provides': 'dv',
    'infer': 'infer_stochastic_DV',
    'arguments': OrderedDict((
        ('localization_error',  ('-e', dict(type=float, help='localization precision (see also sigma; default is 0.03)'))),
        ('diffusivity_prior',   ('-d', dict(type=float, help='prior on the diffusivity'))),
        ('potential_prior',     ('-v', dict(type=float, help='prior on the potential energy'))),
        ('jeffreys_prior',      ('-j', dict(action='store_true', help="Jeffreys' prior"))),
        ('time_prior',          ('-t', dict(type=float, help='prior on the temporal variations of the diffusivity and potential energy (multiplies with `diffusivity_prior` and `potential_prior`'))),
        ('diffusivity_time_prior',('--time-d', dict(type=float, help='prior on the temporal variations of the diffusivity'))),
        ('potential_time_prior',('--time-v', dict(type=float, help='prior on the temporal variations of the potential'))),
        ('min_diffusivity',     dict(type=float, help='minimum diffusivity value allowed')),
        ('max_iter',            dict(type=int, help='maximum number of iterations')),
        ('compatibility',       ('-c', '--inferencemap', '--compatible',
                                dict(action='store_true', help='InferenceMAP compatible'))),
        ('gradient',            ('--grad', dict(help="spatial gradient implementation; any of 'grad1', 'gradn'"))),
        ('grad_epsilon',        dict(args=('--eps', '--epsilon'), kwargs=dict(type=float, help='if defined, every spatial gradient component can recruit all of the neighbours, minus those at a projected distance less than this value'), translate=True)),
        ('grad_selection_angle',('-a', dict(type=float, help='top angle of the selection hypercone for neighbours in the spatial gradient calculation (1= pi radians; if not -c, default is: {})'.format(default_selection_angle)))),
        ('rgrad',               dict(help="local spatial variation; any of 'delta0' (highly recommended), 'delta1'")),
        ('export_centers',      dict(action='store_true')),
        ('verbose',             ()))),
        #('region_size',         ('-s', dict(type=int, help='radius of the regions, in number of adjacency steps'))))),
    'cell_sampling': 'group'}


module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.DEBUG)
_console = logging.StreamHandler()
_console.setFormatter(logging.Formatter('%(message)s'))
module_logger.addHandler(_console)


class LocalDV(DV):
    __slots__ = ('regions','_diffusivity_time_prior','_potential_time_prior',
            'prior_delay','_n_calls','_undefined_grad','_undefined_time_derivative',
            '_update_undefined_grad','_update_undefined_time_derivative','_logger')

    def __init__(self, diffusivity, potential,
        diffusivity_spatial_prior=None, potential_spatial_prior=None,
        diffusivity_time_prior=None, potential_time_prior=None,
        minimum_diffusivity=None, positive_diffusivity=None, prior_include=None,
        regions=None, prior_delay=None, logger=True):
        # positive_diffusivity is for backward compatibility
        DV.__init__(self, diffusivity, potential, diffusivity_spatial_prior, potential_spatial_prior,
            minimum_diffusivity, positive_diffusivity, prior_include)
        self.regions = regions
        self._diffusivity_time_prior = diffusivity_time_prior
        self._potential_time_prior = potential_time_prior
        self.prior_delay = prior_delay
        self._n_calls = 0.
        self._undefined_grad = set()
        self._undefined_time_derivative = set()
        self._update_undefined_grad = set()
        self._update_undefined_time_derivative = set()
        self._logger = logger

    @property
    def verbose(self):
        return self._logger is not None

    @verbose.setter
    def verbose(self, b):
        if b:
            if not self.verbose:
                self._logger = True
        else:
            if self.verbose:
                self._logger = None

    def region(self, i):
        return self.regions[i]

    def indices(self, cell_ids):
        if isinstance(cell_ids, (int, np.int_)):
            return np.array([ cell_ids, int(self.combined.size / 2) + cell_ids ])
        cell_ids = np.array(cell_ids)
        cell_ids.sort()
        return np.concatenate((cell_ids, int(self.combined.size / 2) + cell_ids))
    def diffusivity_indices(self, cell_ids):
        if isinstance(cell_ids, (int, np.int_)):
            cell_ids = [cell_ids]
        return np.array(cell_ids)
    def potential_indices(self, cell_ids):
        if isinstance(cell_ids, (int, np.int_)):
            cell_ids = [cell_ids]
        return int(self.combined.size / 2) + np.array(cell_ids)

    def potential_prior(self, i):
        if self.prior_delay:
            if self._n_calls < self.prior_delay:
                prior = None
            else:
                prior = DV.potential_prior(self, i)
            if self._diffusivity_prior:
                self._n_calls += .5
            else:
                self._n_calls += 1.
        else:
                prior = DV.potential_prior(self, i)
        return prior

    def potential_spatial_prior(self, i):
        return self.potential_prior(i)

    def potential_time_prior(self, i):
        return self._potential_time_prior

    def diffusivity_prior(self, i):
        if self.prior_delay:
            if self._n_calls < self.prior_delay:
                prior = None
            else:
                prior = DV.diffusivity_prior(self, i)
            if self._potential_prior:
                self._n_calls += .5
            else:
                self._n_calls += 1.
        else:
                prior = DV.diffusivity_prior(self, i)
        return prior

    def diffusivity_spatial_prior(self, i):
        return self.diffusivity_prior(i)

    def diffusivity_time_prior(self, i):
        return self._diffusivity_time_prior

    @property
    def logger(self):
        if self._logger is True:
            self._logger = module_logger
        return self._logger

    @logger.setter
    def logger(self, l):
        self._logger = l

    def undefined_grad(self, i, feature=''):
        if self.verbose and i not in self._undefined_grad:
            self._undefined_grad.add(i)
            self._update_undefined_grad.add(i)
            self.logger.debug('grad{}({}) is not defined'.format(feature, i))

    def undefined_time_derivative(self, i, feature=''):
        if self.verbose and i not in self._undefined_time_derivative:
            self._undefined_time_derivative.add(i)
            self._update_undefined_time_derivative.add(i)
            self.logger.debug('d{}({})/dt failed'.format(feature, i))

    def pop_workspace_update(self):
        try:
            return self._update_undefined_grad, self._update_undefined_time_derivative
        finally:
            self._update_undefined_grad, self._update_undefined_time_derivative = set(), set()

    def push_workspace_update(self, update):
        undefined_grad, undefined_time_derivative = update
        if isinstance(undefined_grad, set):
            self._undefined_grad.update(undefined_grad)
            self._undefined_time_derivative.update(undefined_time_derivative)

parallel.abc.WorkspaceExtension.register(LocalDV)


def make_regions(cells, index, reverse_index, size=1):
    A = cells.adjacency
    regions = []
    for i in index:
        if isinstance(i, int):
            j = set([i])
        else:
            j = set([i.tolist()])
        j_inner = set()
        for k in range(size):
            j_outer = j - j_inner
            j_inner = j
            for l in j_outer:
                neighbours = A.indices[A.indptr[l]:A.indptr[l+1]] # NOT cells.neighbours(l)
                j |= set(neighbours.tolist())
        j = reverse_index[list(j)]
        regions.append(j)
    return regions


def lookup_space_cells(cells):
    A = cells.temporal_adjacency
    available = { i for i in cells }
    space_cells = []
    while available:
        space_cell = set()
        front_cells = { available.pop() }
        while front_cells:
            more_cells = set()
            for i in front_cells:
                more_cells |= set(A.indices[A.indptr[i]:A.indptr[i+1]])
            space_cell |= front_cells
            if more_cells:
                front_cells = more_cells - space_cell
            else:
                break
        available -= space_cell
        space_cells.append(list(space_cell))
    return space_cells


def local_dv_neg_posterior(j, x, dv, cells, sigma2, jeffreys_prior,
    dt_mean, index, reverse_index, grad_kwargs,
    posterior_info=None, iter_num=None, verbose=False):
    """
    """

    # extract `D` and `V`
    #dv.update(x)
    #D = dv.D # slow(?)
    #V = dv.V
    #Dj = D[j]
    Dj = x[j]
    if np.any(np.isnan(Dj)):
        raise ValueError('D is nan')
    V = x[int(x.size/2):]
    #

    noise_dt = sigma2

    # for all cell
    i = index[j]
    cell = cells[i]
    n = len(cell) # number of translocations

    # spatial gradient of the local potential energy
    gradV = cells.grad(i, V, reverse_index, **grad_kwargs)
    #print('{}\t{}\t{}\t{}\t{}\t{}'.format(i+1,D[j], V[j], -gradV[0], -gradV[1], result))
    #print('{}\t{}\t{}'.format(i+1, *gradV))
    if gradV is None or np.any(np.isnan(gradV)):
        dv.undefined_grad(i, 'V')
        gradV = np.zeros(cell.dim)

    # various posterior terms
    #print(cell.dt)
    D_dt = Dj * cell.dt
    denominator = 4. * (D_dt + noise_dt)
    if np.any(denominator <= 0):
        raise ValueError('undefined posterior; local diffusion value: %s', Dj)
    dr_minus_drift = cell.dr + np.outer(D_dt, gradV)
    # non-directional squared displacement
    ndsd = np.sum(dr_minus_drift * dr_minus_drift, axis=1)
    raw_posterior = n * log(pi) + np.sum(np.log(denominator)) + np.sum(ndsd / denominator)

    if np.isnan(raw_posterior):
        raise ValueError('undefined posterior; local diffusion value: %s', Dj)

    # priors
    standard_priors, time_priors = 0., 0.
    V_prior = dv.potential_spatial_prior(j)
    if V_prior:
        deltaV = cells.local_variation(i, V, reverse_index, **grad_kwargs)
        if deltaV is not None:
            standard_priors += V_prior * cells.grad_sum(i, deltaV * deltaV, reverse_index)
    D_prior = dv.diffusivity_spatial_prior(j)
    if D_prior:
        D = x[:int(x.size/2)]
        # spatial gradient of the local diffusivity
        deltaD = cells.local_variation(i, D, reverse_index, **grad_kwargs)
        if deltaD is not None:
            # `grad_sum` memoizes and can be called several times at no extra cost
            standard_priors += D_prior * cells.grad_sum(i, deltaD * deltaD, reverse_index)
    #print('{}\t{}\t{}'.format(i+1, D[j], result))
    if jeffreys_prior:
        if Dj <= 0:
            raise ValueError('non positive diffusivity')
        standard_priors += jeffreys_prior * 2. * np.log(Dj * dt_mean[j] + sigma2) - np.log(Dj)

    D_time_prior = dv.diffusivity_time_prior(i)
    if D_time_prior:
        if not D_prior:
            D = x[:int(x.size/2)]
        dDdt = cells.temporal_variation(i, D, reverse_index)
        if dDdt is None:
            dv.undefined_time_derivative(i, 'D')
        else:
            # assume fixed-duration time window
            time_priors += D_time_prior * np.sum(dDdt * dDdt)
    V_time_prior = dv.potential_time_prior(i)
    if V_time_prior:
        dVdt = cells.temporal_variation(i, V, reverse_index)
        if dVdt is None:
            dv.undefined_time_derivative(i, 'V')
        else:
            time_priors += V_time_prior * np.sum(dVdt * dVdt)

    priors = standard_priors + time_priors
    result = raw_posterior + priors
    #if 1e6 < np.max(np.abs(gradV)):
    #    print((i, raw_posterior, standard_priors, Dj, V[j], n, gradV, [V[_j] for _j in cells.neighbours(i)]))

    if verbose:
        dv.logger.debug((i, raw_posterior, standard_priors, time_priors))
    if posterior_info is not None:
        if iter_num is None:
            info = [i, raw_posterior, result]
        else:
            info = [iter_num, i, raw_posterior, result]
        posterior_info.append(info)

    return result

def _local_dv_neg_posterior(*args, **kwargs):
    try:
        return local_dv_neg_posterior(*args, **kwargs)
    except ValueError:
        return np.inf


def infer_stochastic_DV(cells,
    diffusivity_spatial_prior=None, potential_spatial_prior=None,
    diffusivity_time_prior=None, potential_time_prior=None,
    jeffreys_prior=False, min_diffusivity=None, max_iter=None,
    compatibility=False,
    export_centers=False, verbose=True, superlocal=True, stochastic=True,
    D0=None, V0=None, x0=None, rgrad=None, debug=False, fulltime=False,
    diffusion_prior=None, diffusion_spatial_prior=None, diffusion_time_prior=None,
    prior_delay=None, return_struct=False, posterior_max_count=None,# deprecated
    diffusivity_prior=None, potential_prior=None, time_prior=None,
    allow_negative_potential=False,
    **kwargs):
    """
    Arguments:

        ...

        time_prior (float or tuple): one of two regularization coefficients that
            penalize the temporal derivative of diffusivity and potential energy;
            the first coefficient applies to diffusivity (and is multiplied by
            `diffusivity_prior`) and the second applies to potential energy (and
            is multiplied by `potential_prior`).

        diffusivity_spatial_prior/diffusion_spatial_prior: alias for `diffusivity_prior`.

        diffusivity_time_prior/diffusion_time_prior: penalizes the temporal derivative
            of diffusivity; this coefficient does NOT multiply with `diffusivity_prior`.

        potential_spatial_prior: alias for `potential_prior`.

        potential_time_prior: penalizes the temporal derivative of potential;
            this coefficient does NOT multiply with `potential_prior`.

        stochastic (bool): if ``False``, emulate standard *DV* with sparse gradient.

        D0 (float or ndarray): initial diffusivity value(s).

        V0 (float or ndarray): initial potential energy value(s).

        x0 (ndarray): initial parameter vector [diffusivities, potentials].

        rgrad (str): either 'grad'/'grad1', 'gradn' (none recommended), 'delta'/'delta0'
            or 'delta1';
            see the corresponding functions in module :mod:`~tramway.inference.gradient`.

        debug (bool or sequence of str): any subset of {'ncalls', 'f_history',
            'df_history', 'projg_history', 'error', 'diagnoses'}.

        xref (ndarray): reference parameter vector [diffusivities, potentials];
            required to compute the 'error' debug variable.

        diagnosis (callable): function of the iteration number, current component and
            candidate updated component;
            required to compute the 'diagnoses' debug variable.

        prior_delay/return_struct/posterior_max_count: all deprecated.

        allow_negative_potential (bool): do not offset the whole potential map towards
            all-positive values.

        ...

    See also :func:`~tramway.inference.optimization.minimize_sparse_bfgs`.
    """

    # initial values
    if stochastic and not (superlocal or potential_prior or potential_spatial_prior):
        raise ValueError('spatial regularization is required for the potential energy')
    localization_error = cells.get_localization_error(kwargs, 0.03, True)
    index, reverse_index, n, dt_mean, D_initial, _min_diffusivity, D_bounds, border = \
        smooth_infer_init(cells, min_diffusivity=min_diffusivity, jeffreys_prior=jeffreys_prior,
        sigma2=localization_error)
    # V initial values
    if x0 is None:
        if V0 is None:
            try:
                if compatibility:
                    raise Exception # skip to the except block
                volume = [ cells[i].volume for i in index ]
            except:
                V_initial = -np.log(n / np.max(n))
            else:
                density = n / np.array([ np.inf if (v is None or np.isnan(v)) else v for v in volume ])
                try:
                    density[density == 0] = np.min(density[0 < density])
                except ValueError:
                    raise ValueError('no data in bounded domains; null density everywhere')
                V_initial = np.log(np.max(density)) - np.log(density)
    else:
        #warn('`x0` is deprecated; please use `D0` and `V0` instead', DeprecationWarning)
        if x0.size != 2 * D_initial.size:
            raise ValueError('wrong size for x0')
        D_initial, V_initial = x0[:int(x0.size/2)], x0[int(x0.size/2):]
    if D0 is not None:
        if np.isscalar(D0):
            D_initial[...] = D0
        elif D0.size == D_initial.size:
            D_initial = D0
        else:
            raise ValueError('wrong size for D0')
    if V0 is not None:
        if np.isscalar(V0):
            # V0 may be given as an integer
            V_initial = np.full(D_initial.size, V0, dtype=D_initial.dtype)
        elif V0.size == D_initial.size:
            V_initial = V0
        else:
            raise ValueError('wrong size for V0')

    if diffusivity_prior is None:
        diffusivity_prior = diffusion_prior
    if diffusivity_spatial_prior is None:
        if diffusion_spatial_prior is None:
            diffusivity_spatial_prior = diffusivity_prior
        else:
            diffusivity_spatial_prior = diffusion_spatial_prior
    if potential_spatial_prior is None:
        potential_spatial_prior = potential_prior
    if not isinstance(time_prior, (tuple, list)):
        time_prior = (time_prior, time_prior)
    if diffusivity_time_prior is None:
        diffusivity_time_prior = diffusion_time_prior
    if diffusivity_time_prior is None and not (time_prior[0] is None or diffusivity_spatial_prior is None):
        diffusivity_time_prior = time_prior[0] * diffusivity_spatial_prior
    if potential_time_prior is None and not (time_prior[1] is None or potential_spatial_prior is None):
        potential_time_prior = time_prior[1] * potential_spatial_prior
    dv = LocalDV(D_initial, V_initial, diffusivity_spatial_prior, potential_spatial_prior,
        diffusivity_time_prior, potential_time_prior, min_diffusivity, prior_delay=prior_delay)
    if verbose:
        logger = dv.logger
    else:
        dv.logger = None
        logger = module_logger
    if posterior_max_count:
        warn('`posterior_max_count` is deprecated', RuntimeWarning)
    posterior_info = None # parallelization makes this inoperant

    # gradient options
    grad_kwargs = get_grad_kwargs(kwargs)
    # bounds
    if np.isscalar(V0) and V0 == 0:
        V_bounds = [(None, None)] * V_initial.size
    else:
        V_bounds = [(0., None)] * V_initial.size
    #if min_diffusivity is not None:
    #    assert np.all(min_diffusivity < D_initial)
    bounds = D_bounds + V_bounds

    # posterior function input arguments
    if jeffreys_prior is True:
        jeffreys_prior = 1.
    args = (dv, cells, localization_error, jeffreys_prior, dt_mean,
        index, reverse_index, grad_kwargs, posterior_info)

    m = len(index)
    if not stochastic:
        y0 = sum( local_dv_neg_posterior(j, dv.combined, *args[:-1]) for j in range(m) )
        if verbose:
            dv.logger.info('At X0\tactual posterior= {}'.format(y0))
        #args = args + (y0,)

    # keyword arguments to `minimize_sparse_bfgs`
    sbfgs_kwargs = dict(kwargs)

    # cell groups (a given cell + its neighbours)
    dv.regions = make_regions(cells, index, reverse_index)

    # covariates and subspaces
    if stochastic:
        component = m
        covariate = dv.region
        #covariate = lambda i: np.unique(np.concatenate([ dv.region(_i) for _i in dv.region(i) ]))
        descent_subspace = None
        if fulltime:
            space_cells = lookup_space_cells(cells)
            component = len(space_cells)
            def gradient_subspace(i):
                return space_cells[i]
        elif superlocal:
            gradient_subspace = dv.indices
        else:
            #def gradient_subspace(i):
            #    return np.r_[dv.diffusivity_indices(i), dv.potential_indices(dv.region(i))]
            def gradient_subspace(i):
                return dv.indices(dv.region(i))
        if 'independent_components' not in sbfgs_kwargs:
            sbfgs_kwargs['independent_components'] = True
        if 'memory' not in sbfgs_kwargs:
            sbfgs_kwargs['memory'] = None
    else:
        if superlocal:
            superlocal = False
            #raise ValueError('`stochastic=False` and `superlocal=True` are incompatible')
        component = lambda k: 0
        covariate = lambda i: range(m)
        descent_subspace = None
        gradient_subspace = None
        def col2rows(j):
            i = j % m
            return dv.region(i)
        if 'gradient_covariate' not in sbfgs_kwargs:
            sbfgs_kwargs['gradient_covariate'] = col2rows
        sbfgs_kwargs['worker_count'] = 0

    #if os.name == 'nt':
    #    if sbfgs_kwargs.get('worker_count', None):
    #        logger.warning('multiprocessing may break on Windows')
    #    else:
    #        sbfgs_kwargs['worker_count'] = 0

    # other arguments
    if verbose:
        sbfgs_kwargs['verbose'] = verbose
        if 1 < verbose:
            args = args + (None, True)
    if max_iter:
        sbfgs_kwargs['max_iter'] = max_iter
    if bounds is not None:
        sbfgs_kwargs['bounds'] = bounds
    if 'eps' not in sbfgs_kwargs:
        sbfgs_kwargs['eps'] = 1. # beware: not the `tramway.inference.grad` option
    sbfgs_kwargs['newton'] = newton = sbfgs_kwargs.get('newton', True)
    if newton:
        default_step_max = 2.
        default_wolfe = (.5, None)
    else:
        default_step_max = .5
        default_wolfe = (.1, None)
    if True:#stochastic:
        sbfgs_kwargs['ls_wolfe'] = sbfgs_kwargs.get('ls_wolfe', default_wolfe)
        sbfgs_kwargs['ls_step_max'] = sbfgs_kwargs.get('ls_step_max', default_step_max)
    if 'ls_armijo_max' not in sbfgs_kwargs:
        sbfgs_kwargs['ls_armijo_max'] = 5
    #ls_step_max_decay = sbfgs_kwargs.get('ls_step_max_decay', None)
    #if ls_step_max_decay:
    #    sbfgs_kwargs['ls_step_max_decay'] /= float(m)
    if 'ftol' not in sbfgs_kwargs:
        sbfgs_kwargs['ftol'] = 1e-3
        sbfgs_kwargs['low_df_rate'] = 1
    if 'gtol' not in sbfgs_kwargs:
        sbfgs_kwargs['gtol'] = None
    if debug:
        debug_all = {'ncalls': 'ncalls',
                'f_history': 'f',
                'df_history': 'df',
                'projg_history': 'projg',
                'error': 'err',
                'diagnoses': 'diagnosis'}
        if debug is True:
            sbfgs_kwargs['returns'] = 'all'
        else:
            if isinstance(debug, str):
                debug = (debug,)
            sbfgs_kwargs['returns'] = { debug_all[attr] for attr in debug }
    else:
        sbfgs_kwargs['returns'] = set()

    assert not np.any(np.isnan(dv.combined))
    assert not np.any(np.isinf(dv.combined))

    # run the optimization routine
    result = minimize_sparse_bfgs(local_dv_neg_posterior, dv.combined, component, covariate,
            gradient_subspace, descent_subspace, args, **sbfgs_kwargs)
    #if not (result.success or verbose):
    #    warn('{}'.format(result.message), OptimizationWarning)

    # extract the different types of parameters
    dv.update(result.x)
    D, V = dv.D, dv.V
    if np.isscalar(V0) and V0 == 0 and not allow_negative_potential:
        Vmin = np.nanmin(V)
        if Vmin < 0:
            V -= Vmin
    DVF = pd.DataFrame(np.stack((D, V), axis=1), index=index,
        columns=['diffusivity', 'potential'])

    # derivate the forces
    index_, F = [], []
    for i in index:
        gradV = cells.grad(i, V, reverse_index, **grad_kwargs)
        if gradV is not None:
            index_.append(i)
            F.append(-gradV)
    if F:
        F = pd.DataFrame(np.stack(F, axis=0), index=index_,
            columns=[ 'force ' + col for col in cells.space_cols ])
    else:
        dv.logger.warning('not any cell is suitable for evaluating the local force')
        F = pd.DataFrame(np.zeros((0, len(cells.space_cols)), dtype=V.dtype),
            columns=[ 'force ' + col for col in cells.space_cols ])
    DVF = DVF.join(F)

    # add extra information if required
    if export_centers:
        xy = np.vstack([ cells[i].center for i in index ])
        DVF = DVF.join(pd.DataFrame(xy, index=index,
            columns=cells.space_cols))
        #DVF.to_csv('results.csv', sep='\t')

    info = {}
    # format the posteriors
    if posterior_info:
        cols = ['cell', 'fit', 'total']
        if len(posterior_info[0]) == 4:
            cols = ['iter'] + cols
        posterior_info = pd.DataFrame(np.array(posterior_info), columns=cols)
        info['posterior_info'] = posterior_info

    if return_struct:
        # cannot be rwa-stored
        info['result'] = result
    else:
        attrs = ['resolution', 'niter']
        if debug:
            if debug is True:
                attrs += [(val, key) for key, val in debug_all.items()]
            else:
                attrs += [(debug_all[attr], attr) for attr in debug]
        for src_attr in attrs:
            if isinstance(src_attr, str):
                dest_attr = src_attr
            else:
                src_attr, dest_attr = src_attr
            val = getattr(result, src_attr)
            if val is not None:
                info[dest_attr] = val
    return DVF, info

    return DVF, posterior_info




# -*- coding: utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from math import *
import numpy as np
import pandas as pd
from matplotlib.colors import Normalize
import itertools
from copy import deepcopy
import scipy.sparse as sparse
from scipy.spatial.distance import cdist
from ..core import *
from ..tessellation.base import dict_to_sparse, format_cell_index, nearest_cell, Partition, Tessellation
import traceback
from collections import defaultdict


__colors__ = ['darkgreen', 'darkkhaki', 'darkmagenta', 'darkolivegreen', 'darkorange', 'darkorchid', 'darkred', 'darksalmon', 'darkseagreen', 'darkslateblue', 'darkslategray', 'darkviolet', 'deeppink', 'deepskyblue', 'dodgerblue', 'firebrick', 'forestgreen', 'gold', 'goldenrod', 'hotpink', 'indianred', 'indigo', 'lightblue', 'lightcoral', 'lightpink', 'lightsalmon', 'lightseagreen', 'lightskyblue', 'lightsteelblue', 'limegreen', 'maroon', 'mediumaquamarine', 'mediumorchid', 'mediumpurple', 'mediumseagreen', 'mediumslateblue', 'mediumspringgreen', 'mediumturquoise', 'mediumvioletred', 'midnightblue', 'navajowhite', 'navy', 'olive', 'olivedrab', 'orange', 'orangered', 'orchid', 'palegreen', 'paleturquoise', 'palevioletred', 'papayawhip', 'peachpuff', 'peru', 'pink', 'plum', 'powderblue', 'purple', '#663399', 'rosybrown', 'royalblue', 'saddlebrown', 'salmon', 'sandybrown', 'seagreen', 'sienna', 'skyblue', 'slateblue', 'springgreen', 'steelblue', 'tan', 'teal', 'thistle', 'tomato', 'turquoise', 'violet', 'wheat', 'yellowgreen']


def plot_points(cells, min_count=None, style='.', size=8, color=None, axes=None, **kwargs):
    """
    Plot 2D points coloured by associated cell.

    Arguments:

        cells (Partition or FiniteElements):
            full partition

        min_count (int):
            discard cells with less than this number of associated points

        style (str):
            point marker style

        size (int):
            point marker size

        color (str or numpy.ndarray):
            cell colours

    Returns:

        list: handles of the various clouds of points

    Extra keyword arguments are passed to *matplotlib* 's *scatter* or *plot*.
    """
    if axes is None:
        import matplotlib.pyplot as plt
        axes = plt
    if isinstance(cells, np.ndarray):
        points = cells
        label = None
    elif isinstance(cells, Partition):
        points = cells.descriptors(cells.points, asarray=True)
        label = cells.cell_index
        npts = points.shape[0]
        ncells = cells.location_count.size
        # if label is not a single index vector, convert it following
        # tessellation.base.Delaunay.cell_index with `preferred`='force index'.
        merge = nearest_cell(points, cells.tessellation.cell_centers)
        label = format_cell_index(cells.cell_index, format='array', \
            select=merge, shape=(npts, ncells))
        if min_count:
            cell_mask = min_count <= cells.location_count
            label = np.array(label) # copy
            label[np.logical_not(cell_mask[cells.cell_index])] = -1
            #label = cell_mask[cells.cell_index]
    else:#if isinstance(cells, FiniteElements):

        # fully distinct implementation
        handles = []
        if color == 'light' and 'alpha' not in kwargs:
            kwargs['alpha'] = .2
        color = __colors__
        color = list(itertools.islice(itertools.cycle(color), len(cells)))
        for i in cells:
            points = cells[i].origins
            if isinstance(points, pd.DataFrame):
                points = points[['x', 'y']].values
            h = axes.plot(points[:,0], points[:,1], style, color=color[i],
                markersize=size, **kwargs)
            assert not h[1:]
            handles.append(h[0])
        return handles


    # original implementation for the not-FiniteElements case

    if isstructured(points):
        x = points['x']
        y = points['y']
    else:
        x = points[:,0]
        y = points[:,1]

    handles = []

    if label is None or (isinstance(color, str) and len(color)==1):
        if color is None:
            color = 'k'
        elif isinstance(color, (pd.Series, pd.DataFrame)):
            color = np.asarray(color)
        if isinstance(color, np.ndarray):
            cmin, cmax = np.min(color), np.max(color)
            color = (color - cmin) / (cmax - cmin)
            cmap = plt.get_cmap()
            color = [ cmap(c) for c in color ]
        handles.append(axes.scatter(x, y, color=color, marker=style, s=size, **kwargs))
    else:
        L = np.unique(label)
        if color in [None, 'light']:
            if color == 'light' and 'alpha' not in kwargs:
                kwargs['alpha'] = .2
            if 2 < len(L):
                color = __colors__
                color = ['gray'] + \
                    list(itertools.islice(itertools.cycle(color), len(L)))
            elif len(L) == 2:
                color = ['gray', 'k']
            else:   color = 'k'
        for i, l in enumerate(L):
            handles.append(axes.plot(x[label == l], y[label == l],
                style, color=color[i], markersize=size, **kwargs))

    # resize window
    try:
        axes.axis(cells.bounding_box[['x', 'y']].values.flatten('F'))
    except AttributeError:
        pass
    except ValueError:
        print(traceback.format_exc())

    return handles


def plot_voronoi(cells, labels=None, color=None, style='-', centroid_style='g+', negative=None,
        linewidth=1, fallback_color='gray', verbose=True, axes=None):
    """
    Voronoi plot.

    Arguments:

        cells (Partition):
            full partition

        labels (numpy.ndarray):
            numerical labels for cell adjacency relationship

        color (str):
            single-character colours in a string, e.g. 'rrrbgy'

        style (str):
            line style

        centroid_style (str):
            marker style of the cell centers

        negative (any):
            if ``None``, do not plot edges corresponding to negative adjacency labels

        linewidth (int):
            line width

        fallback_color (str):
            colour for missing edges

        verbose (bool):
            print message about missing edges

        axes (matplotlib.axes.Axes):
            figure axes

    Returns:

        tuple: list of handles of the plotted edges,
            handle of the plotted centroids
    """
    if axes is None:
        import matplotlib.pyplot as plt
        axes = plt
    vertices = cells.tessellation.vertices
    labels, color = _graph_theme(cells.tessellation, labels, color, negative)
    try:
        color += 'w'
    except TypeError:
        if not isinstance(color, list):
            color = [color]
        color.append('w')
    try:
        special_edges = cells.tessellation.candidate_edges
        #points = cells.descriptors(cells.points, asarray=True)
    except:
        special_edges = {}
    c = 0 # if cells.tessellation.adjacency_label is None
    fallback_clr = 'k'
    edge_handles, centroid_handle = [], None
    # plot voronoi
    #plt.plot(vertices[:,0], vertices[:,1], 'b+')
    if cells.tessellation.adjacency_label is not None or special_edges:
        n_cells = cells.tessellation._cell_centers.shape[0]
        n_vertices = vertices.shape[0]
        cell_vertex = dict_to_sparse(cells.tessellation.cell_vertices, \
                shape=(n_cells, n_vertices)).tocsc()
        adjacency = cells.tessellation.cell_adjacency.tocsr()
    Av = sparse.tril(cells.tessellation.vertex_adjacency, format='coo')
    d2 = np.sum((vertices[Av.row[0]] - vertices[Av.col[0]])**2)
    for u, v in zip(Av.row, Av.col):
        x, y = vertices[[u, v]].T
        if cells.tessellation.adjacency_label is not None or special_edges:
            try:
                a, b = set(cell_vertex[:,u].indices) & set(cell_vertex[:,v].indices)
                # adjacency may contain explicit zeros
                js = list(adjacency.indices[adjacency.indptr[a]:adjacency.indptr[a+1]])
                # js.index(b) will fail if a and b are not adjacent
                edge_ix = adjacency.data[adjacency.indptr[a]+js.index(b)]
            except (ValueError, IndexError):
                if verbose:
                    print("vertices {} and {} do not match with a ridge".format(u, v))
                    print(traceback.format_exc())
                #continue
                c = -1
            else:
                if cells.tessellation.adjacency_label is not None:
                    try:
                        c = labels.index(cells.tessellation.adjacency_label[edge_ix])
                    except ValueError:
                        continue
        if 0 <= c:
            try:
                _clr = color[c]
            except IndexError:
                import warnings
                warnings.warn("too few colours: '{}'; index {:d} out of range".format(color, c), RuntimeWarning)
                _clr = fallback_color
        else:
            _clr = fallback_color
        h = axes.plot(x, y, style, color=_clr, linewidth=linewidth)
        assert not h[1:]
        edge_handles.append(h[0])

        # extra debug steps
        if special_edges and edge_ix in special_edges:
            #i, j, ii, jj = special_edges[edge_ix]
            #try:
            #       i = points[cells.cell_index == i][ii]
            #       j = points[cells.cell_index == j][jj]
            #except IndexError as e:
            #       print(e)
            #       continue
            i, j = special_edges[edge_ix]
            x_, y_ = zip(i, j)
            axes.plot(x_, y_, 'c-')
            x_, y_ = (i + j) / 2
            axes.text(x_, y_, str(edge_ix), \
                horizontalalignment='center', verticalalignment='center')

    centroids = cells.tessellation.cell_centers
    # plot cell centers
    if centroid_style:
        h = axes.plot(centroids[:,0], centroids[:,1], centroid_style)
        assert not h[1:]
        centroid_handle = h[0]

    # resize window
    try:
        axes.axis(cells.bounding_box[['x', 'y']].values.flatten('F'))
    except AttributeError:
        pass
    except ValueError:
        print(traceback.format_exc())

    return edge_handles, centroid_handle


def plot_delaunay(cells, labels=None, color=None, style='-', centroid_style='g+', negative=None,
        axes=None, linewidth=1, individual=False, fallback_color='gray'):
    """
    Delaunay plot.

    Arguments:

        cells (Partition):
            full partition

        labels (numpy.ndarray):
            numerical labels for cell adjacency relationship

        color (str):
            single-character colours in a string, e.g. 'rrrbgy'

        style (str):
            line style

        centroid_style (str):
            marker style of the cell centers

        negative (any):
            if ``None``, do not plot edges corresponding to negative adjacency labels;
            if '*voronoi*', plot the corresponding Voronoi edge instead, for edges with
            negative labels

        axes (matplotlib.axes.Axes):
            figure axes

        linewidth (int):
            line width

        individual (bool):
            plot each edge independently; this generates a lot of handles and takes time

        fallback_color (str):
            colour for unexpected labels

    Returns:

        tuple: list of handles of the plotted edges,
            handle of the plotted centroids
    """
    if axes is None:
        import matplotlib.pyplot as plt
        axes = plt
    try:
        tessellation = cells.tessellation
    except AttributeError:
        tessellation = cells

    vertices = tessellation.cell_centers
    if negative == 'voronoi':
        voronoi = tessellation.cell_vertices

    labels, color = _graph_theme(tessellation, labels, color, negative)

    # if asymetric, can be either triu or tril
    A = sparse.triu(tessellation.cell_adjacency, format='coo')
    I, J, K = A.row, A.col, A.data
    if not I.size:
        A = sparse.tril(tessellation.cell_adjacency, format='coo')
        I, J, K = A.row, A.col, A.data

    if not individual:
        by_color = defaultdict(list)
    edge_handles, centroid_handle = [], None # handles

    # plot delaunay
    for i, j, k in zip(I, J, K):
        x, y = zip(vertices[i], vertices[j])
        if labels is None:
            c = 0
        else:
            label = tessellation.adjacency_label[k]
            try:
                c = labels.index(label)
            except ValueError:
                continue
            if label <= 0:
                if negative == 'voronoi':
                    try:
                        vert_ids = set(tessellation.cell_vertices.get(i, [])) & set(tessellation.cell_vertices.get(j, []))
                        x, y = voronoi[vert_ids].T
                    except ValueError:
                        continue
        if individual:
            h = axes.plot(x, y, style, color=color[c], linewidth=linewidth)
            assert not h[1:]
            edge_handles.append(h)
        else:
            by_color[c].append((x, y))

    if not individual:
        if not color[1:]:
            _clr = color[0]
        for c in by_color:
            xy = by_color[c]
            X = np.zeros((len(xy) * 3,))
            Y = np.empty((len(xy) * 3,))
            Y[:] = np.nan
            i = 0
            for x, y in xy:
                I = slice(i*3, i*3+2)
                X[I], Y[I] = x, y
                i += 1
            if color[1:]:
                try:
                    _clr = color[c]
                except IndexError:
                    import warnings
                    warnings.warn('too few specified colours; at least {:d} needed'.format(c), RuntimeWarning)
                    _clr = fallback_color
            h = axes.plot(X, Y, style, color=_clr, linewidth=linewidth)
            assert not h[1:]
            edge_handles.append(h[0])

    # plot cell centers
    if centroid_style:
        h = axes.plot(vertices[:,0], vertices[:,1], centroid_style)
        assert not h[1:]
        centroid_handle = h[0]

    # resize window
    try:
        axes.axis(cells.bounding_box[['x', 'y']].values.flatten('F'))
    except AttributeError:
        pass
    except ValueError:
        print(traceback.format_exc())

    return edge_handles, centroid_handle


def _graph_theme(tess, labels, color, negative):
    if tess.adjacency_label is None:
        if not color:
            color = 'r'
    else:
        if labels is None:
            labels = np.unique(tess.adjacency_label).tolist()
    if labels is not None:
        if negative is None:
            labels = [ l for l in labels if 0 < l ]
            nnp = 0
        else:
            nnp = len([ l for l in labels if l <= 0 ]) # number of non-positive labels
    if not color:
        neg_color = 'cym'
        pos_color = 'rgb'
        labels.sort()
        color = neg_color[:nnp] + pos_color
    return (labels, color)


def plot_distributed(cells, vertex_color='g', vertex_style='x', edge_color='r',
        arrow_size=5, arrow_color='y'):
    """
    Plot a :class:`~tramway.inference.base.FiniteElements` object as a mesh.

    Arguments:

        cells (tramway.inference.base.FiniteElements):
            mesh prepared for the inference

        vertex_color (str): colour for the cell centers

        vertex_style (str): marker style for the cell centers

        edge_color (str): colour of the edge between adjacent cells

        arrow_size (int): size of the arrows along the edges

        arrow_color (str): colour of the arrows along the edges

    Returns:

        tuple: list of handles of the edges,
            handle of the cell centers (vertices),
            list of handles of the arrow segments

    `plot_distributed` is similar to `plot_delaunay` but takes a :class:`~tramway.inference.base.FiniteElements` object instead.
    """
    import matplotlib.pyplot as plt

    centers = np.vstack([ cells[i].center for i in cells ])
    _min, _max = np.min(centers), np.max(centers)

    plot_arrows = arrow_size is not None and 0 < arrow_size and arrow_color is not None
    if plot_arrows:
        arrow_size = float(arrow_size) * 1e-3 * (_max - _min)
        half_base = arrow_size / sqrt(5.)

    edges, arrows, vertices = [], [], None
    plotted = defaultdict(list)
    for i in cells:
        x = cells[i].center
        for j in cells.adjacency[i].indices:
            y = cells[j].center

            # edge
            if j not in plotted[i]:
                h = plt.plot([x[0], y[0]], [x[1], y[1]], edge_color+'-')
                assert not h[1:]
                edges.append(h[0])
                plotted[i].append(j)

            # arrow
            if plot_arrows:
                dr = y - x
                top = x + .6667 * dr
                dr /= sqrt(np.sum(dr * dr))
                bottom = top - arrow_size * dr
                n = np.array([-dr[1], dr[0]])
                left, right = bottom + half_base * n, bottom - half_base * n
                h = plt.plot([left[0], top[0], right[0]], [left[1], top[1], right[1]], arrow_color+'-')
                assert not h[1:]
                arrows.append(h[0])

    # cell centers
    plot_vertices = vertex_color and vertex_style
    if plot_vertices:
        h = plt.plot(centers[:,0], centers[:,1], vertex_color+vertex_style)
        assert not h[1:]
        vertices = h[0]

    return edges, vertices, arrows


def plot_cell_indices(cells, font_size=12, shift_indices=False, axes=None, **kwargs):
    """
    Plot cell indices at the cell centers.

    Arguments:

        cells (Partition or Tessellation or FiniteElements):
            tessellation

        font_size (int):
            alias for `fontsize`

        shift_indices (bool):
            first cell is numbered 1 instead of 0

        axes (matplotlib.axes.Axes):
            figure axes; *new in 0.6*

    Returns:

        list: handles of the individual text elements

    Trailing keyword arguments are passed to :func:`~matplotlib.pyplot.text`.
    """
    if axes is None:
        import matplotlib.pyplot as plt
        axes = plt
    kwargs['fontsize'] = kwargs.get('fontsize', font_size)
    handles = []
    if isinstance(cells, Partition):
        cells = cells.tessellation
    # common plotting logic
    def text(x, y, i):
        if np.isnan(x) or np.isnan(y):
            import warnings
            warnings.warn('nan coordinate', RuntimeWarning)
            return
        elif np.isinf(x) or np.isinf(y):
            import warnings
            warnings.warn('inf coordinate', RuntimeWarning)
            return
        h = axes.text(x, y, str(i+1 if shift_indices else i), **kwargs)
        handles.append(h)
    # Partition and Tessellation
    if isinstance(cells, Tessellation):
        i = 0
        for x,y in cells.cell_centers:
            text(x, y, i)
            i += 1
    # FiniteElements
    else:#if isinstance(cells, FiniteElements):
        for i in cells:
            x,y = cells[i].center
            text(x, y, i)
    return handles


def plot_indices(*args, **kwargs):
    """
    Alias for :func:`plot_cell_indices`.

    *plot_cell_indices* was formerly named *plot_indices*.
    """
    return plot_cell_indices(*args, **kwargs)


__all__ = ['__colors__', 'plot_points', 'plot_voronoi', 'plot_delaunay', 'plot_distributed', 'plot_indices', 'plot_cell_indices']




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from abc import *

class Analyses(metaclass=ABCMeta):
    @property
    @abstractmethod
    def data(self):
        pass
    @data.setter
    @abstractmethod
    def data(self, d):
        pass
    @property
    @abstractmethod
    def artefact(self):
        pass
    @artefact.setter
    @abstractmethod
    def artefact(self, a):
        pass
    @property
    @abstractmethod
    def metadata(self):
        pass
    @metadata.setter
    @abstractmethod
    def metadata(self, d):
        pass
    @property
    @abstractmethod
    def instances(self):
        pass
    @property
    @abstractmethod
    def comments(self):
        pass
    @property
    @abstractmethod
    def labels(self):
        pass
    @abstractmethod
    def autoindex(self, pattern=None):
        pass
    @abstractmethod
    def add(self, analysis, label=None, comment=None, raw=False):
        pass

__all__ = ['Analyses']





from . import *
from ..attribute import initializer_method

from_sliding_window = initializer_method( TimeInitializer.from_sliding_window )
from_sampling       = initializer_method( TimeInitializer.from_sampling       )




import numpy as np 
import pandas as pd

import sys
import argparse
from   os.path import abspath
import gc
import os
import matplotlib.pyplot as plt
from scipy.optimize import linear_sum_assignment

####################################################################
####################################################################
####################################################################

####################################################################
####################################################################
####################################################################
def load_xyt(path_xyt):


	colonne_names = ['x', 'y','t'] 
	xyt_pandas     = pd.read_csv(path_xyt, header=None,sep='\t', names=colonne_names)
	xyt            = xyt_pandas.values


	return xyt, xyt_pandas
####################################################################
####################################################################
####################################################################
def convert_to_list(xyt):

	t_unique        = np.unique(xyt[:,2])
	n_unique        = t_unique.shape[0]
	movie_per_frame = []
	for i in range(n_unique):

		II = xyt[:,2]==t_unique[i]
		movie_per_frame.append(xyt[II,:])

	return movie_per_frame, n_unique	

####################################################################
####################################################################
####################################################################
def get_dt(movie_per_frame, n_unique):

	ddtt    = []
	dt_theo = movie_per_frame[1][0,2] - movie_per_frame[0][0,2]

	for i in range(n_unique - 1):
		ddtt.append(movie_per_frame[i+1][0,2] - movie_per_frame[i][0,2])

	dt_theo = min(ddtt)
	t_init  = movie_per_frame[0][0,2]
	t_end   = movie_per_frame[n_unique - 1][0,2]

	return dt_theo, t_init, t_end
####################################################################
####################################################################
####################################################################
def get_local_parameters_accessible_elsewhere_in_tramway(dt_theo):

	sigma        = 0.025
	D_high       = 0.3
	length_high  = 2.*np.sqrt(2*D_high*dt_theo)


	return sigma, D_high, length_high
####################################################################
####################################################################
####################################################################
def get_cost_function(number_movie, movie_per_frame):


	x2 = movie_per_frame[number_movie+1][:,0]
	x1 = movie_per_frame[number_movie][:,0]

	y2 = movie_per_frame[number_movie+1][:,1]
	y1 = movie_per_frame[number_movie][:,1]

	dx = x2[:,np.newaxis] - x1
	dy = y2[:,np.newaxis] - y1


	C  = dx**2 + dy**2
	C = C.transpose()

	return C


####################################################################
####################################################################
####################################################################
def correct_cost_function(C,length_high):


	(M,N)            = C.shape
	# cut the links
	l2               = length_high**2
	C[C>l2]          = np.inf
	# positions whith no link possible
	non_inf          = ~np.isinf(C)
	num_col          = np.sum(non_inf,axis=0)
	num_row          = np.sum(non_inf,axis=1)
	## portion that need to be optimised
	row_reduced      = np.where(num_row != 0);
	col_reduced      = np.where(num_col != 0);
	n_row_reduced    = np.size(row_reduced)
	n_col_reduced    = np.size(col_reduced)

	nn               = np.maximum(n_row_reduced,n_col_reduced)

	if (n_row_reduced>1)&(n_col_reduced > 1):  
		CC               = np.squeeze(C[row_reduced,:])
		CC               = np.squeeze(CC[:,col_reduced])
		#square the matrix
		nn               = np.maximum(n_row_reduced,n_col_reduced)
		C_reduced        = np.zeros((nn,nn))
		C_reduced[0:n_row_reduced,0:n_col_reduced] = CC[:,:]
		d_max            = np.amax(C_reduced , where=~np.isinf(C_reduced) , initial=-1)

		# adjust the isze of the matrix to ensure reasonnable assugments
		edge             = np.zeros((nn,nn)) 
		edge[:,:]        = C_reduced[:,:]
		non_inf          = ~np.isinf(C_reduced)
		edge[non_inf]    = 0
		n_add            = correct_deficiencies(edge)
		nn               = nn + n_add
		## corrected matric cost function 

		C_reduced_corrected  = np.ones((nn,nn))*d_max
		C_reduced_corrected[0:n_row_reduced,0:n_col_reduced] =  CC[:,:]
		##elements usefull for final assigment 

		row_reduced      = np.squeeze(np.array(row_reduced))
		col_reduced      = np.squeeze(np.array(col_reduced))

		inf_loc          = np.isinf(C_reduced_corrected)
		C_reduced_corrected[inf_loc] = d_max*1e10
		anomaly          = 0

	elif (n_row_reduced==0)&(n_col_reduced==0):

		C_reduced_corrected = np.zeros(1)
		C_reduced           = np.zeros(1)
		edge                = np.zeros(1)
		d_max               = np.inf
		row_reduced         = np.zeros(1)
		col_reduced         = np.zeros(1)
		n_row_reduced       = 0
		n_col_reduced       = 0
		anomaly             = 2

	elif (n_row_reduced==1):
		#print("here row")
		#print()
		CC               = np.squeeze(C[row_reduced,:])
		CC               = np.squeeze(CC[col_reduced])
		#square the matrix
		nn               = np.maximum(n_row_reduced,n_col_reduced)
		C_reduced        = np.zeros((n_row_reduced,nn))
		#C_reduced        = np.zeros((nn,nn))
		#print((n_row_reduced, n_col_reduced))
		if nn == 1:
			C_reduced[0,0] = CC
		else:
			C_reduced[0,0:n_col_reduced] = CC[:]
			#C_reduced[0:n_row_reduced,0:n_col_reduced] = CC[:]		
			#print(C_reduced)

		d_max            = np.amax(C_reduced , where=~np.isinf(C_reduced) , initial=-1)

		# adjust the isze of the matrix to ensure reasonnable assugments
		edge = []
		## corrected matric cost function 
		if nn==1:
			C_reduced_corrected = d_max
		else:
			C_reduced_corrected  = C_reduced
		
		##elements usefull for final assigment 
		row_reduced      = np.squeeze(np.array(row_reduced))
		col_reduced      = np.squeeze(np.array(col_reduced))
		if nn>1:
			inf_loc          = np.isinf(C_reduced_corrected)
			C_reduced_corrected[inf_loc] = d_max*1e10
		anomaly          = 1

	elif (n_col_reduced==1):
		#print("here col")

		CC               = np.squeeze(C[row_reduced,:])
		CC               = np.squeeze(CC[:,col_reduced])
		#square the matrix
		nn               = np.maximum(n_row_reduced,n_col_reduced)
		C_reduced        = np.zeros((n_row_reduced,n_col_reduced))
		#C_reduced        = np.full((nn, nn), np.inf)
		C_reduced[0:n_row_reduced,0] = CC[:]
		d_max            = np.amax(C_reduced , where=~np.isinf(C_reduced) , initial=-1)

		# adjust the isze of the matrix to ensure reasonnable assugments
		edge = []
		## corrected matric cost function 
		if nn==1:
			C_reduced_corrected = d_max
		else:
			C_reduced_corrected  = C_reduced

		##elements usefull for final assigment 

		row_reduced      = np.squeeze(np.array(row_reduced))
		col_reduced      = np.squeeze(np.array(col_reduced))

		if nn>1:
			inf_loc          = np.isinf(C_reduced_corrected)
			C_reduced_corrected[inf_loc] = d_max*1e10
		anomaly          = 1



	return C_reduced_corrected, C_reduced, edge, d_max, row_reduced, col_reduced,M,N, n_row_reduced, n_col_reduced, anomaly

####################################################################
####################################################################
####################################################################
def get_assigment_matrix_from_reduced_cost(C_reduced_corrected,row_reduced,col_reduced,M,N,n_col_reduced, n_row_reduced,anomaly):
	## most cases
	if (anomaly == 0):

		row_ind, col_ind = linear_sum_assignment(C_reduced_corrected)
		global_row       = []
		global_col       = []
		assingment       = np.zeros((M,N))

		for i in range(len(row_reduced)):
			if col_ind[i] < n_col_reduced:
				coll    = col_reduced[col_ind[i]]
				index   = row_reduced[i]
					
				global_row.append(index)
				global_col.append(coll)
				assingment[index,coll]  = 1 
		global_row = np.array(global_row)
		global_col = np.array(global_col)

	# cases where only one column or one row survived
	elif (anomaly ==1):

		if (n_col_reduced==1)&(n_row_reduced==1):

			row_ind     = 0
			col_ind     = 0
			global_col  = col_reduced
			global_row  = row_reduced
			assingment  = np.zeros((M,N))
			assingment[global_row,global_col] = 1

		elif (n_col_reduced==1):
			row_ind, col_ind = linear_sum_assignment(C_reduced_corrected)

			#row_ind     = 0
			#col_ind     = 0
			global_col  = col_reduced
			global_row  = row_reduced[row_ind]
			assingment  = np.zeros((M,N))
			assingment[global_row,global_col] = 1

		elif(n_row_reduced==1):
			row_ind, col_ind = linear_sum_assignment(C_reduced_corrected)

			#row_ind     = 0
			#col_ind     = 0
			global_col  = col_reduced[col_ind]
			global_row  = row_reduced
			assingment  = np.zeros((M,N))
			assingment[global_row,global_col] = 1

		else:
		
			row_ind, col_ind = linear_sum_assignment(C_reduced_corrected)
			global_row       = []
			global_col       = []
			assingment       = np.zeros((M,N))

			for i in range(len(row_reduced)):
				if col_ind[i] < n_col_reduced:
					coll    = col_reduced[col_ind[i]]
					index   = row_reduced[i]
					
					global_row.append(index)
					global_col.append(coll)
					assingment[index,coll]  = 1 
			global_row = np.array(global_row)
			global_col = np.array(global_col)



	# no link can be made evrything is impossible : zero assigment
	elif anomaly == 2:
		assingment       = np.zeros((M,N))
		global_row       = np.array([])
		global_col       = np.array([])



	return assingment, global_row, global_col

####################################################################
####################################################################
####################################################################
def plot_linking_two_images(movie_per_frame, indice, row_ind, col_ind):
	## red first frame
	## blue second frame

	xyt1 = movie_per_frame[indice]
	xyt2 = movie_per_frame[indice+1]

	plt.scatter(xyt1[:,0],xyt1[:,1], s=5,  c='r')
	plt.scatter(xyt2[:,0],xyt2[:,1], s=10, c='b')

	dx      = np.squeeze( xyt2[col_ind,0] - xyt1[row_ind,0] )
	dy      = np.squeeze( xyt2[col_ind,1] - xyt1[row_ind,1] )
	x_start = np.squeeze( xyt1[row_ind,0] )
	y_start = np.squeeze( xyt1[row_ind,1] )
	#print(x_start, y_start, dx, dy)

	try:
		for idx in range( len(dx) ):
			plt.arrow(x_start[idx],y_start[idx], dx[idx] ,dy[idx] , fc='k', ec='k',head_width=0.5 )
	except TypeError:
			plt.arrow(x_start,y_start, dx ,dy , fc='k', ec='k',head_width=0.5 )




	#plt.axis('scaled')
	plt.axis('scaled')
	plt.show()


	return 1
####################################################################
####################################################################
####################################################################
def plot_linking_from_list(movie_per_frame, indice, liste_row, liste_col ):
	## red first frame
	## blue second frame

	xyt1 = movie_per_frame[indice]
	xyt2 = movie_per_frame[indice+1]

	row_ind  = liste_row[indice]
	col_ind  = liste_col[indice]	

	plt.scatter(xyt1[:,0],xyt1[:,1], s=5,  c='r')
	plt.scatter(xyt2[:,0],xyt2[:,1], s=10, c='b')

	dx      = np.squeeze( xyt2[col_ind,0] - xyt1[row_ind,0] )
	dy      = np.squeeze( xyt2[col_ind,1] - xyt1[row_ind,1] )
	x_start = np.squeeze( xyt1[row_ind,0] )
	y_start = np.squeeze( xyt1[row_ind,1] )
	#print(x_start, y_start, dx, dy)

	try:
		for idx in range( len(dx) ):
			plt.arrow(x_start[idx],y_start[idx], dx[idx] ,dy[idx] , fc='k', ec='k',head_width=0.5 )
	except TypeError:
			plt.arrow(x_start,y_start, dx ,dy , fc='k', ec='k',head_width=0.5 )




	#plt.axis('scaled')
	plt.axis('scaled')
	plt.show()



	return 1
####################################################################
####################################################################
####################################################################
def plot_linking_3_from_list(movie_per_frame, indice, liste_row, liste_col ):
	## red first frame
	## blue second frame

	xyt1 = movie_per_frame[indice]
	xyt2 = movie_per_frame[indice+1]
	xyt3 = movie_per_frame[indice+2]


	row_ind  = liste_row[indice]
	col_ind  = liste_col[indice]	

	plt.scatter(xyt1[:,0],xyt1[:,1], s=5,  c='r')
	plt.scatter(xyt2[:,0],xyt2[:,1], s=10, c='b')

	dx      = np.squeeze( xyt2[col_ind,0] - xyt1[row_ind,0] )
	dy      = np.squeeze( xyt2[col_ind,1] - xyt1[row_ind,1] )
	x_start = np.squeeze( xyt1[row_ind,0] )
	y_start = np.squeeze( xyt1[row_ind,1] )
	#print(x_start, y_start, dx, dy)

	try:
		for idx in range( len(dx) ):
			plt.arrow(x_start[idx],y_start[idx], dx[idx] ,dy[idx] , fc='k', ec='k',head_width=0.5 )
	except TypeError:
			plt.arrow(x_start,y_start, dx ,dy , fc='k', ec='k',head_width=0.5 )




	#plt.axis('scaled')
	plt.axis('scaled')
	plt.show()



	return 1


####################################################################
####################################################################
####################################################################
def correct_deficiencies(edge):


	(K_edge, L_edge) = edge.shape
	#print((K_edge, L_edge))
	row              = np.zeros((K_edge,1))
	col              = np.zeros((K_edge,1))
	edge_eff         = np.zeros((K_edge,K_edge))



	for ii in range(K_edge):
		for jj in range(K_edge):
			if (edge[ii,jj]==0)&(row[ii]==0)&(col[jj]==0):
				edge_eff[ii,jj] = 1
				row[ii]=1
				col[jj]=1
				break


	row = np.zeros((K_edge,1))
	indicateur_1 = 1

	while indicateur_1:
		row_loc = -1
		col_loc = -1
		indicateur_2 = 1
		ii      = 0
		jj      = 0
		#print(indicateur_1)
		while indicateur_2:
			if (edge[ii,jj]==0)&(row[ii]==0)&(col[jj]==0):
				row_loc = ii
				col_loc = jj
				indicateur_2 = 0

			jj = jj + 1
			if (jj>=K_edge):
 #				print("here jj\n")
				jj=0
				ii=ii+1
				#print(ii)
			if (ii>=K_edge):
 #				print("here ii \n")
				indicateur_2=0

		if row_loc ==-1:
			#print("here")
			indicateur_1 = 0
		else:
			#print((row_loc,col_loc))
			edge_eff[row_loc,col_loc] = 2
			III = edge_eff[row_loc,:] ==1
			#print(III)
			if (np.sum(III) !=0):
				row[row_loc]  = 1
				col_col2      = edge_eff[row_loc,:]==1
				col[col_col2] = 0
			else:
				#print("here\n")
				indicateur_1 = 0
		#print(indicateur_1)




	n_add = K_edge - np.sum(col) - np.sum(row)
	return int(n_add)


####################################################################
####################################################################
####################################################################
#def give_optimal_assigment(C, length_high):
	##. cutoff
#	(M,N)            = C.shape
#	l2               = length_high**2
	#print(np.sqrt(l2))
#	C[C>l2]          = np.inf
	## where there is a possibility to do a match
#	non_inf          = ~np.isinf(C)
#	num_col          = np.sum(non_inf,axis=0)
#	num_row          = np.sum(non_inf,axis=1)

#	row_reduced      = np.where(num_row != 0);
#	col_reduced      = np.where(num_col != 0);
#	CC               = np.squeeze(C[row_reduced,:])

#	row_reduced      = np.squeeze(np.array(row_reduced))
	#print(row_reduced.shape)
	## reduced matrix with non-inf value
#	CC[np.isinf(CC)] = np.amax(CC , where=~np.isinf(CC) , initial=-1)

	## optimal assingment
#	row_ind, col_ind = linear_sum_assignment(CC)
	#print(row_ind.shape)
#	index            = row_reduced[row_ind]
#	assingment       = np.zeros((M,N))
#	assingment[index,col_ind]     = 1

#	return assingment, index, col_ind
####################################################################
####################################################################
####################################################################



####################################################################
####################################################################
####################################################################






####################################################################
####################################################################
####################################################################





####################################################################
####################################################################
####################################################################





####################################################################
####################################################################
####################################################################






#import warnings as _warnings
#_warnings.filterwarnings('ignore', category=FutureWarning, module='h5py', lineno=36)
#_warnings.filterwarnings('default', category=DeprecationWarning)
#_warnings.filterwarnings('default', category=PendingDeprecationWarning)

from . import core

__all__ = ['core']




# -*- coding: utf-8 -*-

# Copyright © 2018, Alexander Serov


import logging
import warnings

# from multiprocessing import Pool
import numpy as np
import scipy.optimize
from numpy.linalg import norm

from .calculate_marginalized_integral import (calculate_integral_ratio,
                                              calculate_marginalized_integral)
from .calculate_posteriors import get_lambda_MAP
from .convenience_functions import n_pi_func
from .convenience_functions import p as pow
from .stopwatch import stopwatch

# from scipy.special import gammainc


try:
    from tqdm import trange  # for graphical estimate of the progress
except ImportError:
    trange = range


class NaNInputError(ValueError):
    pass


# def calculate_bayes_factors_for_cells(cells, loc_error, dim=2, B_threshold=10, verbose=True):
#     """Calculate Bayes factors for an iterable ensemble of cells."""
#
#     check_dimensionality(dim)
#
#     # M = len(cells)
#     lg_Bs, forces, min_ns = [], [], []
#     for i, cell in enumerate(cells):
#         lg_Bs[i], forces[i], min_ns[i] = calculate_bayes_factors_for_one_cell(
#             cell, loc_error, dim=dim, B_threshold=B_threshold, verbose=verbose)
#         cell.lg_B, cell.force, cell.min_n = lg_Bs[i], forces[i], min_ns[i]
#
#     return [lg_Bs, forces, min_ns]


def calculate_bayes_factors_for_one_cell(cell, loc_error, dim=2, B_threshold=10, verbose=True):
    """Calculate Bayes factors for one cell.
    The results are saved to cell properties.
    """

    check_dimensionality(dim)

    try:
        cell.lg_B, cell.force, cell.min_n = _calculate_one_bayes_factor(
            zeta_t=cell.zeta_total,
            zeta_sp=cell.zeta_spurious,
            n=cell.n,
            V=cell.V,
            V_pi=cell.V_prior,
            loc_error=loc_error,
            dim=dim,
            bl_need_min_n=True,
            B_threshold=B_threshold)
    except NaNInputError:
        cell.lg_B, cell.force, cell.min_n = [np.nan] * 3
        raise NaNInputError()

    # cell.lamb_MAP = get_lambda_MAP(zeta_t=cell.zeta_total, zeta_sp=cell.zeta_spurious)
    return [cell.lg_B, cell.force, cell.min_n]


def calculate_bayes_factors(zeta_ts, zeta_sps, ns, Vs, Vs_pi, loc_error, dim=2, B_threshold=10, verbose=True):
    """
    Calculate the Bayes factor for a set of bins given a uniform localization error.

    Input:
    zeta_ts --- signal-to-noise ratios for the total force = dx_mean / sqrt(var(dx)) in bins. Size: M x 2,
    zeta_sps --- signal-to-noise ratios for the spurious force = grad(D) / sqrt(var(dx)) in bins. Size: M x 2,
    ns --- number of jumps in each bin. Size: M x 1,
    Vs --- jump variance in bins = E((dx - dx_mean) ** 2). Size: M x 1,
    Vs_pi --- jump variance in all other bins relative to the current bin. Size: M x 1,
    loc_error --- localization error. Same units as variance. Set to 0 if localization error can be ignored;
    dim --- dimensionality of the problem;
    B_threshold --- the values of Bayes factor for thresholding.

    Output:
    Bs, forces, min_ns

    lg_Bs --- log_10 Bayes factor values in the bins. Size: M x 1,
    forces --- Returns 1 if there is strong evidence for the presence of a conservative forces,
    -1 for strong evidence for 	a spurious force, and 0 if the is not enough evidence. Size: M x 1;
    min_ns --- minimum number of data points in a bin to support the currently favored model at the required evidence level. Size: M x 1.

    Notation:
    M --- number of bins.
    """
    check_dimensionality(dim)

    if verbose:
        def _trange(x): return trange(x, desc='Bayes factor calculation')
    else:
        _trange = range

    # Convert to numpy
    vars = [zeta_ts, zeta_sps, Vs, Vs_pi, ns]
    vars = map(np.asarray, vars)

    # Check that the 2nd dimension has size 2
    if np.shape(zeta_ts)[1] != 2 or np.shape(zeta_sps)[1] != 2:
        raise ValueError("zeta_ts and zeta_sps must be matrices of size (M x 2)")

    M = len(ns)

    # Calculate
    lg_Bs = np.zeros_like(ns) * np.nan
    forces = np.zeros_like(ns) * np.nan
    min_ns = np.zeros_like(ns, dtype=int) * np.nan
    nan_cells_list = []
    with stopwatch("Bayes factor calculation", verbose):
        for i in _trange(M):
            lg_Bs[i], forces[i], min_ns[i] = _calculate_one_bayes_factor(
                zeta_ts[i, :], zeta_sps[i, :], ns[i], Vs[i], Vs_pi[i], loc_error, dim)
            try:
                lg_Bs[i], forces[i], min_ns[i] = _calculate_one_bayes_factor(
                    zeta_ts[i, :], zeta_sps[i, :], ns[i], Vs[i], Vs_pi[i], loc_error, dim)
            except NaNInputError:
                nan_cells_list.append(i)

        # Report error if any
        try:
            logging.warn(
                "A NaN value was present in the input parameters for the following cells: {nan_cells_list}.\nBayes factor calculations were skipped for them".format(nan_cells_list=nan_cells_list))
        except NameError:
            pass

        return [lg_Bs, forces, min_ns]


def _calculate_one_bayes_factor(zeta_t, zeta_sp, n, V, V_pi, loc_error, dim, B_threshold=10, bl_need_min_n=True):
    """Calculate the Bayes factor for one bin."""

    test = check_for_nan(zeta_t, zeta_sp, n, V, V_pi, loc_error)
    if test != 'ok':
        raise NaNInputError()
        # logging.warning(
        #     f'>>A {test} value is present in the input parameters for _calculate_one_bayes_factor.\nSkipping Bayes factor calculation for the current bin.\nCall parameters: zeta_t={zeta_t}, zeta_sp={zeta_sp}, n={n}, V={V}, V_pi={V_pi}, loc_error={loc_error}<<')
        # return [np.nan] * 3

    # # Check if None is present
    # if any(var is None for var in [zeta_t, zeta_sp, n, V, V_pi, loc_error]):
    #     logging.info('None values encountered in cell. Skipping cell.')
    #     return [np.nan] * 3

    # Parameter combinations
    n_pi = n_pi_func(dim)
    p = pow(n, dim)
    u = V_pi / V
    v = 1.0 + n_pi / n * u
    eta = np.sqrt(n_pi / (n + n_pi))

    if loc_error > 0:
        rel_loc_error = n * V / (4 * loc_error)
    else:
        rel_loc_error = np.inf

    def upstairs(l):
        return v + eta**2 * norm(zeta_t - l * zeta_sp)**2

    def downstairs(l):
        return v + norm(zeta_t - l * zeta_sp)**2

    # Identify break points
    with np.errstate(invalid='ignore', divide='ignore'):
        bps = set(zeta_t / zeta_sp)
        bps = [bp for bp in bps if bp >= 0 and bp <= 1]

    ln_B = calculate_integral_ratio(arg_func_up=upstairs, arg_func_down=downstairs,
                                    pow_up=p, pow_down=p, v=v,
                                    rel_loc_error=rel_loc_error,
                                    break_points=bps, lamb='marg')
    lg_B = ln_B / np.log(10) + dim * np.log10(eta)

    if bl_need_min_n:
        min_n = calculate_minimal_n(
            zeta_t=zeta_t, zeta_sp=zeta_sp, n0=n, V=V, V_pi=V_pi, loc_error=loc_error, dim=dim, B_threshold=B_threshold)
    else:
        min_n = np.nan

    # Threshold
    force = 1 * (lg_B >= np.log10(B_threshold)) - 1 * (lg_B <= -np.log10(B_threshold))

    return [lg_B, force, min_n]


def calculate_minimal_n(zeta_t, zeta_sp, n0, V, V_pi, loc_error, dim=2, B_threshold=10):
    """
    Calculate the minimal number of jumps per bin needed to obtain strong evidence for the active force or the spurious force model.

    Input:
    zeta_t, zeta_sp - -- vectors of length 2 for one bin
    n0 - -- initial number of jumps(evidence already available). This way the "next" strong evidence can be found, i.e. the minimal number of data points to support the current conclusion

    Output:
    min_n - -- minimal number of jumps to obtain strong evidence for the conservative force model.
    Return - 1 if unable to find the min_n
    """

    test = check_for_nan(zeta_t, zeta_sp, n0, V, V_pi, loc_error)
    if test != 'ok':
        logging.warning(
            '>>A {test} value is present in the input parameters for calculate_minimal_n.\nSkipping minimal n calculation for the current bin.\nCall parameters: zeta_t={zeta_t}, zeta_sp={zeta_sp}, n0={n0}, V={V}, V_pi={V_pi}, loc_error={loc_error}<<'.format(
                test=test, zeta_t=zeta_t, zeta_sp=zeta_sp, n0=n0, V=V, V_pi=V_pi, loc_error=loc_error))
        return np.nan

    if np.isnan(n0):
        logging.error(
            'Invalid initial jumps number supplied for minimal n calculation: {}'.format(n0))
        return np.nan
    # Local constants
    increase_factor = 2  # initial search interval increase with each iteration
    max_attempts = 40
    xtol = 1.0
    rtol = 0.001

    lg_B_threshold = np.log10(B_threshold)

    # Define the Bayes factor
    def lg_B(n):
        """A wrapper for the Bayes factor as a function of n."""
        lg_B, _, _ = _calculate_one_bayes_factor(
            zeta_t=zeta_t, zeta_sp=zeta_sp, n=n, V=V, V_pi=V_pi, loc_error=loc_error, dim=dim, bl_need_min_n=False)
        return lg_B

    if abs(lg_B(n0)) >= abs(lg_B_threshold):
        return n0

    # Find the initial search interval
    bl_found = False
    n = n0
    # print(
    #     f'Start min n search. Params: zeta_t={zeta_t}, zeta_sp={zeta_sp}, n={n}, V={V}, V_pi={V_pi}, loc_error={loc_error}')
    for attempt in range(max_attempts):
        n = n0 - 1 + increase_factor ** attempt
        # print(
        #     f'Min_n search. Iteration: {attempt}, n: {n}, lg_B(n): {lg_B(n)}, abs(lg_B_threshold): {abs(lg_B_threshold)}')
        if abs(lg_B(n)) >= abs(lg_B_threshold):
            bl_found = True
            break

    if not bl_found:
        logging.warning(
            "Unable to find the minimal number of data points to provide strong evidence. Search parameters: zeta_t={zeta_t}, zeta_sp={zeta_sp}, n={n}, V={V}, V_pi={V_pi}, loc_error={loc_error}")
        return np.nan

    # Find a more accurate location
    n_interval = [n0, n]
    sign = np.sign(lg_B(n))

    def solve_me(n):
        return lg_B(n) - sign * lg_B_threshold

    min_n = scipy.optimize.brentq(solve_me, n_interval[0], n_interval[1], xtol=xtol, rtol=rtol)
    min_n = np.int(np.ceil(min_n))

    return min_n


def check_dimensionality(dim):
    if dim not in [2]:
        raise ValueError("Bayes factor calculations in {dim}D not supported yet.".format(dim=dim))


def check_for_nan(*args):
    """Return 'nan' if a nan value is present in any of the variables.
    Return 'None' if any of the variables is None.
    Return 'ok' otherwise"""
    for i, var in enumerate(args):
        try:
            if var is None:
                return 'None'
            elif np.any(np.isnan(var)):
                return 'nan'
        except:
            logging.warning('Unable to check for nan value for variable #{i}'.format(i=i))
            return 'unidentified'
    return 'ok'



# import datetime
import logging
import sys
from collections import OrderedDict

import numpy as np

from tramway.tessellation.base import point_adjacency_matrix

from .calculate_bayes_factors import (NaNInputError, calculate_bayes_factors,
                                      calculate_bayes_factors_for_one_cell)
from .group_by_sign import group_by_sign

# The package can be imported by just `import bayes_factors`.
__all__ = ['calculate_bayes_factors', 'calculate_bayes_factors_for_one_cell', 'setup']


if sys.version_info < (3, 5):
    raise RuntimeError("Python 3.5+ is required for calculating Bayes factors")


def _bayes_factor(cells, B_threshold=None, verbose=True, **kwargs):
    if verbose:
        try:
            from tqdm import tqdm
        except:
            logging.warning(
                "Consider installing `tqdm` package (`pip install tqdm`) to see Bayes factors calculation progress.")

            def tqdm(x, desc=None): return x
    else:
        def tqdm(x, desc=None): return x

    # TODO: use the same localization error as for inference
    # input arguments
    # loc_error
    localization_error = cells.get_localization_error(kwargs)
    if localization_error is None:
        raise RuntimeError("Localization error must be specified for calculating Bayes factors")
    # B_threshold
    if B_threshold is not None:
        kwargs['B_threshold'] = B_threshold
    # verbose
    if verbose is not None:
        kwargs['verbose'] = verbose

    # iterate over the cells
    nan_cells_list = []
    for key in tqdm(cells):
        try:
            calculate_bayes_factors_for_one_cell(cells[key], localization_error, **kwargs)
        except NaNInputError:
            nan_cells_list.append(key)

    # Report error if any
    if len(nan_cells_list) > 0:
        logging.warn(
            "A NaN value was present in the input parameters for the following cells: {nan_cells_list}.\nBayes factor calculations were skipped for them".format(nan_cells_list=nan_cells_list))

    # Group cells by Bayes factor
    group_by_sign(cells=cells, tqdm=tqdm, **kwargs)


setup = {
    'name': 'bayes_factor',
    'infer': '_bayes_factor',
    'arguments': OrderedDict((
        ('localization_error', ('-e', dict(type=float, help='localization error (same units as the variance)'))),
        ('B_threshold', ('-b', dict(type=float, help='values of Bayes factor for thresholding'))),
        ('verbose', ()),
    )),
    # List of variables that the module returns as cell properties, e.g. cell.lg_B
    'returns': ['lg_B', 'force', 'min_n', 'groups', 'group_lg_B', 'group_forces'],
}



# -*- coding: utf-8 -*-

# Copyright © 2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from . import stochastic_dv as sdv
import numpy as np
import logging


setup = dict(sdv.setup) # copy

setup.update({'name': 'semi.stochastic.dv',
    'infer': 'infer_DV'})


module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.DEBUG)
_console = logging.StreamHandler()
_console.setFormatter(logging.Formatter('%(message)s'))
module_logger.addHandler(_console)


def infer_DV(cells, *args, **kwargs):
    """
    See also :func:`~tramway.inference.stochastic_dv.infer_stochastic_DV`.
    """

    ## stochastic start

    n = len(cells)

    kwargs['stochastic'] = True

    max_iter = kwargs.pop('max_iter', None)
    kwargs['max_iter'] = kwargs.pop('init_max_iter', 300) * n

    #try:
    #    ftol = kwargs['ftol']
    #except KeyError:
    #    ftol = None
    #else:
    #    kwargs['ftol'] = ftol / n

    verbose = kwargs.get('verbose', False)
    if verbose:
        module_logger.debug('======== stochastic optimization ========\n')

    DV0, info0 = sdv.infer_stochastic_DV(cells, *args, **kwargs)

    D0 = DV0['diffusivity'].values
    V0 = DV0['potential'].values

    if info0['resolution'] == 'interrupted':
        return DV0, info0

    ## non-stochastic fine-tuning

    kwargs['stochastic'] = False
    kwargs['D0'] = D0
    kwargs['V0'] = V0

    try:
        ls_step_max = kwargs['ls_step_max']
    except KeyError:
        pass
    else:
        if isinstance(ls_step_max, np.ndarray):
            kwargs['ls_step_max'] = ls_step_max.max()

    if max_iter is None:
        kwargs.pop('max_iter')
    else:
        kwargs['max_iter'] = max_iter

    #if ftol is not None:
    #    kwargs['ftol'] = ftol
    try:
        ftol = kwargs['ftol']
    except KeyError:
        pass
    else:
        kwargs['ftol'] = ftol * n

    if verbose:
        module_logger.debug('\n======= non-stochastic fine-tuning =======\n')
    try:

        DV, info = sdv.infer_stochastic_DV(cells, *args, **kwargs)

    except KeyboardInterrupt:
        DV, info = DV0, info0
    else:
        info['init'] = info0

    return DV, info




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .proxy import *
# for overloading :meth:`tessellate`
from ..artefact import analysis
from copy import deepcopy


from tramway.tessellation.grid import RegularMesh

class Squares(FreeResolutionTesseller):
    """
    Wraps :class:`tramway.tessellation.grid.RegularMesh`.
    """
    __slots__ = ()
    def __init__(self, **kwargs):
        TessellerProxy.__init__(self, RegularMesh, **kwargs)
        self.alg_name = 'grid'
    def _reset_kwargs(self):
        TessellerProxy._reset_kwargs(self)
        self._init_kwargs.update(dict(
                min_probability = None,
                avg_probability = None,
                max_probability = None,
                min_distance = None,
                avg_distance = None,
                avg_location_count = None,
                ))

    #scaler          = proxy_property('scaler', 'attr')
    lower_bound     = proxy_property('lower_bound', 'attr')
    """ *ndarray* or *Series*: Scaled lower space bounds """
    upper_bound     = proxy_property('upper_bound', 'attr')
    """ *ndarray* or *Series*: Scaled upper space bounds """
    count_per_dim   = proxy_property('count_per_dim',   'attr')
    """ *int*: Number of squares per space dimension """
    #min_probability = proxy_property('min_probability', 'attr')
    #""" *float*: (not used) """
    avg_probability = proxy_property('avg_probability', 'attr')
    """ *float*: Average probability for a point to fall within any given square """
    #max_probability = proxy_property('max_probability', 'attr')
    #""" *float*: (not used) """
    min_distance    = proxy_property('min_distance',    'attr')
    """
    *float*:
        Minimum distance between adjacent square centers or, equivalently, minimum square size;
        ignored if :attr:`avg_distance` is defined
    """
    avg_distance    = proxy_property('avg_distance',    'attr')
    """
    *float*:
        Average distance between adjacent square centers or, equivalently, average square size;
        ignored if :attr:`avg_probability` is defined
    """


from tramway.tessellation.hexagon import HexagonalMesh

class Hexagons(FreeResolutionTesseller):
    """
    Wraps :class:`tramway.tessellation.hexagon.HexagonalMesh`.
    """
    __slots__ = ()
    def __init__(self, **kwargs):
        TessellerProxy.__init__(self, HexagonalMesh, **kwargs)
        self.alg_name = 'hexagon'
    def _reset_kwargs(self):
        TessellerProxy._reset_kwargs(self)
        self._tessellate_kwargs = dict(
                lower_bound = None,
                upper_bound = None,
                )
        self._init_kwargs.update(dict(
                tilt = 0.,
                min_probability = None,
                avg_probability = None,
                max_probability = None,
                min_distance = None,
                avg_distance = None,
                avg_location_count = None,
                ))

    #scaler          = proxy_property('scaler', 'attr')
    tilt            = proxy_property('tilt',    'attr')
    r"""
    *float*:
        Angle of the "main" axis, in :math:`\frac{\pi}{6}` radians;
        0= the main axis is the x-axis; 1= the main axis is the y-axis
    """
    #min_probability = proxy_property('min_probability', 'attr')
    #""" *float*: (not used) """
    avg_probability = proxy_property('avg_probability', 'attr')
    """ *float*: Average probability for a point to fall within any given hexagon """
    #max_probability = proxy_property('max_probability', 'attr')
    #""" *float*: (not used) """
    min_distance    = proxy_property('min_distance',    'attr')
    """
    *float*:
        Minimum distance between adjacent hexagon centers or, equivalently,
        minimum diameter of the inscribed circle;
        ignored if :attr:`avg_distance` is defined
    """
    avg_distance    = proxy_property('avg_distance',    'attr')
    """
    *float*:
        Average distance between adjacent hexagon centers or, equivalently,
        average diameter of the inscribed circle;
        ignored if :attr:`avg_probability` is defined
    """
    lower_bound     = proxy_property('lower_bound', 'tessellate')
    """ *ndarray* or *Series*: Scaled lower space bounds """
    upper_bound     = proxy_property('upper_bound', 'tessellate')
    """ *ndarray* or *Series*: Scaled upper space bounds """


from tramway.tessellation.kmeans import KMeansMesh

class KMeans(FreeResolutionTesseller):
    """
    Wraps :class:`tramway.tessellation.kmeans.KMeansMesh`.
    """
    __slots__ = ()
    def __init__(self, **kwargs):
        TessellerProxy.__init__(self, KMeansMesh, **kwargs)
        self.alg_name = 'kmeans'
    def _reset_kwargs(self):
        TessellerProxy._reset_kwargs(self)
        self._tessellate_kwargs = dict(
                tol          = 1e-7,
                prune        = False, # differs from default
                plot         = False,
                avg_distance = None,
                )
        self._init_kwargs.update(dict(
                min_distance = None,
                avg_probability = None,
                avg_location_count = None,
                ))

    #scaler          = proxy_property('scaler', 'attr')
    initial         = proxy_property('initial', 'attr')
    """
    *str*:
        Any of :const:`'grid'` (default), :const:`'random'` and :const:`'center'`;
        :const:`'grid'` uses :class:`Squares` to initialize the cell centers;
        :const:`'random'` randomly samples the initial cell centers;
        :const:`'center'` is similar to :const:`'random'` but confines the random cell
        centers within a tiny area in the middle
    """
    avg_probability = proxy_property('avg_probability', 'attr')
    """ *float*: Average probability for a point to fall within any given cell """
    tol             = proxy_property('tol', 'tessellate')
    """ *float*: Error tolerance; passed as `thresh` to :func:`scipy.cluster.vq.kmeans` """
    prune           = proxy_property('prune',   'tessellate')
    """
    *bool* or *float*:
        Maximum edge length for the Voronoi graph, relative to the median edge length;
        if :const:`True`, defaults to :const:`2.5`;
        the edges that exceed the derived threshold value are pruned
    """
    #plot            = proxy_property('plot',    'tessellate')
    #_min_distance   = proxy_property('_min_distance',   'attr')
    #""" *float*: Scaled minimum distance between adjacent cell centers """
    min_distance    = proxy_property('min_distance',    '__init__')
    """
    *float*:
        Minimum distance between adjacent cell centers;
        affects the :const:`'grid'` initialization only
    """
    avg_distance    = proxy_property('avg_distance',    'tessellate')
    """
    *float*:
        Average distance between adjacent cell centers;
        affects the :const:`'grid'` initialization only
    """


from tramway.tessellation.gwr import GasMesh

class GWR(TessellerProxy):
    """
    Wraps :class:`tramway.tessellation.gwr.GasMesh`.
    """
    __slots__ = ()
    def __init__(self, **kwargs):
        TessellerProxy.__init__(self, GasMesh, **kwargs)
        self.alg_name = 'gwr'
        self.min_distance = 0 # differs from default
    def _reset_kwargs(self):
        TessellerProxy._reset_kwargs(self)
        self._tessellate_kwargs = dict(
                batch_size         = 10000,
                tau                = 333.,
                trust              = 1.,
                lifetime           = .05, # differs from default
                pass_count         = (),
                residual_factor    = .7,
                error_count_tol    = 5e-3,
                min_growth         = 1e-4,
                collapse_tol       = .01,
                stopping_criterion = 0,
                verbose            = False,
                alpha_risk         = 1e-15,
                complete_delaunay  = True, # differs from default
                topology           = None, # differs from default
                )
        self._init_kwargs.update(dict(
                min_distance = None, # default overridden in __init__
                avg_distance = None,
                max_distance = None,
                min_probability = None,
                min_location_count = None,
                ))

    #scaler             = proxy_property('scaler',  'attr')
    min_probability    = proxy_property('min_probability',  'attr')
    """ *float*: Minimum probability for a point to fall within any given cell """
    _min_distance      = proxy_property('_min_distance',    'attr')
    _avg_distance      = proxy_property('_avg_distance',    'attr')
    _max_distance      = proxy_property('_max_distance',    'attr')
    batch_size         = proxy_property('batch_size',   'tessellate')
    """
    *int*:
        Number of points per epoch;
        this merely affects the node merging mechanism triggered between epochs
    """
    tau                = proxy_property('tau',  'tessellate')
    """
    *float* or *(float, float)*:
        Habituation scaling factor, in number of iterations,
        for the nearest and second nearest nodes respectively
    """
    trust              = proxy_property('trust',    'tessellate')
    r"""
    *float*:
        Confidence in outstanding points to become new nodes;
        the new node is set as :math:`\frac{(1-\rm{trust})*w + (1+\rm{trust})*\eta}{2}`;
        the original GWR algorithm corresponds to ``trust = 0``
    """
    lifetime           = proxy_property('lifetime', 'tessellate')
    """
    *float*:
        Edge lifetime in number of iterations;
        taken as absolute, if greater than 1;
        otherwise relative to the number of nodes in the graph
    """
    pass_count         = proxy_property('pass_count',   'tessellate')
    """
    *float* or *(float, float)*:
        Number of points to sample (with replacement), as a multiple of the
        total size of the data;
        a pair of values denotes the lower and upper bounds on the number
        of points;
        a single value is interpreted as the lower bound,
        and the upper bound is then set equal to ``2 * pass_count``
    """
    residual_factor    = proxy_property('residual_factor',  'tessellate')
    """
    *float*:
        Maximum residual for each sample point, relative to :attr:`_max_distance`
    """
    error_count_tol    = proxy_property('error_count_tol',  'tessellate')
    """
    *float*:
        Maximum rate of errors as defined by :attr:`residual_factor` and the batch size;
        as long as the rate does not fall below this threshold value, :meth:`tessellate`
        keeps on sampling batches
    """
    min_growth         = proxy_property('min_growth',   'tessellate')
    """
    *float*:
        Minimum relative increase in the number of nodes
    """
    collapse_tol       = proxy_property('collapse_tol', 'tessellate')
    """
    *float*:
        Maximum ratio of the number of collapsed nodes over the total
        number of nodes
    """
    #stopping_criterion = proxy_property('stopping_criterion',   'tessellate')
    #""" *int*: (deprecated) """
    verbose            = proxy_property('verbose',  'tessellate')
    """ *bool*: Verbose mode; default is :const:`False` """
    alpha_risk         = proxy_property('alpha_risk',   'tessellate')
    """
    *float*:
        Threshold p-value in the comparison between distance distributions
        to determine whether the edges that are not in the Delaunay graph
        are still valid
    """
    complete_delaunay  = proxy_property('complete_delaunay',    'tessellate')
    """
    *bool*:
        Use the Delaunay graph instead, for determining the cell adjacency;
        default is :const:`True`
    """
    topology           = proxy_property('topology', 'tessellate')
    """
    *str*:
        Any of :const:`'approximate density'` and :const:`'displacement length'`
        (default for trajectory/translocation data)
    """
    @analysis
    def tessellate(self, spt_dataframe):
        if not isinstance(self._tesseller, tuple):
            self.calibrate(spt_dataframe)
        tesseller = deepcopy(self.tesseller)
        # modified part below
        delta_columns = [ 'd'+col for col in self.colnames ]
        if all([ col in spt_dataframe.columns for col in delta_columns ]):
            data = (spt_dataframe[self.colnames], spt_dataframe[delta_columns])
            if self.topology is None:
                self.topology = 'displacement length'
        else:
            data = spt_dataframe[self.colnames]
        tesseller.tessellate(data, **self._tessellate_kwargs)
        # modified part above
        if self.post_processing.initialized:
            tesseller = self.post_processing.post_process(tesseller, spt_dataframe[self.colnames])
        #
        return tesseller
    @property
    def resolution(self):
        if self.ref_distance is None:
            return None
        else:
            return 2. * self.ref_distance
    @resolution.setter
    def resolution(self, res):
        if res is None:
            self.ref_distance = None
        else:
            self.ref_distance = .5 * res
            try:
                self.min_probability = 0
            except AttributeError:
                pass


__all__ = ['Squares', 'Hexagons', 'KMeans', 'GWR']




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


class AnalysisBrowser(object):
    __slots__ = ('_analyses', '_path', '_subtree')
    def __init__(self, analyses=None):
        self.analyses = analyses
    @property
    def analyses(self):
        return self._analyses
    @analyses.setter
    def analyses(self, analyses):
        if isinstance(analyses, str):
            import os.path
            if os.path.isfile(analyses):
                from tramway.core.hdf5.store import load_rwa
                analyses = load_rwa(analyses, lazy=True)
        self._analyses = analyses
        self._path = None
        self._subtree = None
    @property
    def subtree(self):
        if self._subtree is None and self.analyses is not None:
            self._subtree = self.analyses
            for label in self.path():
                self._subtree = self._subtree[label]
        return self._subtree
    @subtree.setter
    def subtree(self, analyses):
        self._subtree = analyses
    @property
    def artefact(self):
        return None if self.subtree is None else self._subtree.artefact
    def path(self):
        if self._path is not None:
            yield from self._path
    def labels(self):
        if self.subtree is not None:
            yield from self.subtree.labels
    def select_child(self, label):
        self.subtree = self.subtree[label]
        if self._path is None:
            self._path = [label]
        else:
            self._path.append(label)
    def select_parent(self):
        if self._path is None:
            self._subtree = self.analyses
        else:
            child = self.analyses
            path, label = [], None
            it = self.path()
            while True:
                try:
                    child_label = next(it)
                except StopIteration:
                    break
                self._subtree = child
                if label is not None:
                    path.append(label)
                label = child_label
                child = child[label]
            self._path = path




# -*- coding: utf-8 -*-

# Copyright © 2018, Institut Pasteur
#   Contributor: Jean-Baptiste Masson

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import numpy as np
import pandas as pd
import random
#import pylab
#import math
import collections


############################################################
############################################################


class generate_trajectories:


    def __init__(self, lambda_ = 1, sigma_ = 0.4, sigma_noise_    = 0.010,  dt_  = 25e-3,
       n_short_ = 1000, N_mean_ = 7, n_trajs_  = 100, L_ = 0.5, density_  = 2. , D0_   = 0.1,
       amplitude_D_ = 0.025, amplitude_V_ = 0,
       mode_D="linear", mode_V="potential_force", mode_gamma="equilibrium", name_file_out="trajectories.txt", verbose=True):

       self.lambda_         = lambda_
       self.sigma_          = sigma_
       self.sigma_noise_    = sigma_noise_
       self.dt_             = dt_
       self.dt_space_       = 10 * self.dt_
       self.n_short_        = n_short_
       self.N_mean_         = N_mean_
       self.n_trajs_        = n_trajs_
       self.L_              = L_
       self.density_        = density_
       self.D0_             = D0_
       self.gamma_          = 1 / D0_
       self.n_total_size_max_   = n_trajs_ * 3 * N_mean_
       self.Nb_tot_loc_     = density_ * L_ * L_
       self.N_duration_tot_ = np.round(n_trajs_/self.Nb_tot_loc_)
       self.dt_short_       = dt_ / n_short_

       self.amplitude_D_    = amplitude_D_
       self.amplitude_V_    = amplitude_V_


       self.mode_D          = mode_D
       self.mode_V          = mode_V
       self.mode_gamma      = mode_gamma

       self.name_file_out   = name_file_out

       self.verbose = verbose

############################################################
############################################################
    def  random_start(self):
         L_ = self.L_
         x = -L_ + np.random.random()*2*L_
         y = -L_ + np.random.random()*2*L_
         return x , y
############################################################
############################################################
    def traj_duration(self):
        N_mean_ = self.N_mean_
        N_duration_ = -N_mean_*np.log(np.random.random())
        return N_duration_
############################################################
############################################################
    def noisify(self):
        sigma_noise_  = self.sigma_noise_
        x_            = self.x_
        y_            = self.y_

        nn_            = x_.size

        sigma1_       = np.random.randn(nn_)
        sigma2_       = np.random.randn(nn_)

        x_            = x_ + sigma_noise_*sigma1_
        y_            = y_ + sigma_noise_*sigma2_

        self.x_      = x_
        self.y_      = y_
############################################################
############################################################
    def diffusion_gaussian_trap(self, x, y ):

        r2_           = x*x + y*y;
        D0_           = self.D0_
        sigma_        = self.sigma_
        amplitude_D_  = self.amplitude_D_
        sigma_2_      = sigma_ * sigma_
        D_            = D0_ - amplitude_D_*np.exp(-r2_/(2*sigma_2_))
        grad_D_x_     = amplitude_D_*x/(2*sigma_2_) * np.exp(-r2_/(2*sigma_2_))
        grad_D_y_     = amplitude_D_*y/(2*sigma_2_) * np.exp(-r2_/(2*sigma_2_))

        return D_, grad_D_x_, grad_D_y_
############################################################
############################################################
    def diffusion_gaussian_bump(self, x, y ):

        r2_           = x*x + y*y;
        D0_          = self.D0_
        sigma_       = self.sigma_
        amplitude_D_ = self.amplitude_D_
        sigma_2_     = sigma_ * sigma_
        D_            = D0_ + amplitude_D_*np.exp(-r2_/(2*sigma_2_))
        grad_D_x_     = - amplitude_D_*x/(2*sigma_2_) * np.exp(-r2_/(2*sigma_2_))
        grad_D_y_     = - amplitude_D_*y/(2*sigma_2_) * np.exp(-r2_/(2*sigma_2_))

        return D_, grad_D_x_, grad_D_y_

############################################################
############################################################
    def diffusion_linear(self, x, y):


        D0_          = self.D0_
        amplitude_D_ = self.amplitude_D_
        L_           = self.L_

        D_           = D0_ - amplitude_D_/2 + amplitude_D_/(2*L_)*(x+L_)

        grad_D_x_ = amplitude_D_/(2*L_)
        grad_D_y_ = 0.
        return D_, grad_D_x_, grad_D_y_
############################################################
############################################################
    def potential_trap(self, x, y):


        amplitude_V_ = self.amplitude_V_
        r2_           = x*x + y*y;
        sigma_       = self.sigma_
        sigma_2_     = sigma_ * sigma_

        V_    = 0 -  amplitude_V_*np.exp(-r2_/(2*sigma_2_))
        fx_   = -amplitude_V_*x/(sigma_2_)* np.exp(-r2_/(2*sigma_2_))
        fy_   = -amplitude_V_*y/(sigma_2_)* np.exp(-r2_/(2*sigma_2_))

        return V_, fx_, fy_

############################################################
############################################################
    def potential_drift(self, x, y):

        amplitude_V_ = self.amplitude_V_
        L_           = self.L_

        V_    = 0. -  amplitude_V_/2 + amplitude_V_/(2*L_)*(x+L_)
        fx_   = -amplitude_V_/(2*L_)
        fy_   = 0.

        return V_, fx_, fy_
############################################################
############################################################
    def give_gamma_equilibirum(self, D):

        gamma  = 1/D
        return gamma

############################################################
############################################################
    def give_gamma_fixed_value(self):

        gamma = self.gamma_

        return gamma

############################################################
############################################################
    def give_diffusion_drift_gamma(self, u, v):

        mode_D     = self.mode_D
        mode_V     = self.mode_V
        mode_gamma = self.mode_gamma

        if mode_D == "linear":
            D_, grad_D_x_, grad_D_y_ = self.diffusion_linear(u, v)
        elif mode_D == "gaussian_trap":
            D_, grad_D_x_, grad_D_y_ = self.diffusion_gaussian_trap( u, v )
        elif mode_D == "gaussian_bump":
            D_, grad_D_x_, grad_D_y_ = self.diffusion_gaussian_bump( u, v )

        if mode_V == "potential_force":
            V_, fx_, fy_ = self.potential_trap(u, v)
        elif mode_V == "potential_linear":
            V_, fx_, fy_ = self.potential_drift( u, v)

        if mode_gamma    == "equilibrium":
            gamma_       = self.give_gamma_equilibirum( D_)
        elif  mode_gamma == "fixed":
            gamma_       = self.give_gamma_fixed_value()

        return     D_ , grad_D_x_ , grad_D_y_ , gamma_ , V_ , fx_ , fy_


############################################################
############################################################


    def create_traj(self):
        nb_       = self.nb_
        x_        = self.x_
        y_        = self.y_
        t_        = self.t_
        traj      = pd.DataFrame(collections.OrderedDict([("n", nb_), ("x", x_), ("y", y_), ("t", t_)]))
        self.traj = traj

############################################################
############################################################
    def print_traj(self):
        traj           = self.traj
        name_file_out  = self.name_file_out
        traj.to_csv(name_file_out, sep='\t', index=False, header=False)
############################################################
############################################################
    def update_position(self, u_ , v_ , t_ ):
        dt_short_ = self.dt_short_
        lambda_   = self.lambda_
        D_, grad_D_x_, grad_D_y_, gamma_, V_, fx_, fy_ = self.give_diffusion_drift_gamma( u_, v_)

        u_  = u_ + (fx_/gamma_ + lambda_ *grad_D_x_)*dt_short_ +  np.sqrt(2 * D_ * dt_short_) * np.random.randn()
        v_  = v_ + (fy_/gamma_ + lambda_ *grad_D_y_)*dt_short_ +  np.sqrt(2 * D_ * dt_short_) * np.random.randn()
        t_  = t_ + dt_short_
        return u_, v_, t_
############################################################
############################################################
    def generate_the_actual_trajectories(self):

        n_total_size_max_ = self.n_total_size_max_
        n_trajs_          = self.n_trajs_
        n_short_          = self.n_short_
        dt_space_         = self.dt_space_

        nb_               = []#np.empty(n_total_size_max_, dtype="int")
        x_                = []#np.empty(n_total_size_max_, dtype="float16")
        y_                = []#np.empty(n_total_size_max_, dtype="float16")
        t_                = []#np.empty(n_total_size_max_, dtype="float16")


        indice_ = 0
        t       = 0
        for i in range(1,n_trajs_ + 1):
            if self.verbose:
                print(i)
            t                 = t + dt_space_
            u_, v_            = self.random_start()
            #nb_[indice_]      = i
            #x_[indice_]       = u_
            #y_[indice_]       = v_
            #t_[indice_]       = t
            nb_.append(i)
            x_.append(u_)
            y_.append(v_)
            t_.append(t)
            indice_           = indice_ + 1
            N_duration_       = self.traj_duration()
            N_duration_       = np.round(N_duration_)
            for j in range(1, int( N_duration_ * n_short_ + 1) ):
                #D_ , grad_D_x_ , grad_D_y_ , gamma_ , V_ , fx_ , fy_ = self.give_diffusion_drift_gamma( u_, v_)
                u_, v_, t  = self.update_position( u_ , v_ , t )
                if j%n_short_ == 0:
                    #nb_[indice_]      = i
                    #x_[indice_]       = u_
                    #y_[indice_]       = v_
                    #t_[indice_]       = t
                    nb_.append(i)
                    x_.append(u_)
                    y_.append(v_)
                    t_.append(t)
                    indice_           = indice_ + 1


        #nb_               = nb_[:indice_]
        #x_                = x_[:indice_]
        #y_                = y_[:indice_]
        #t_                = t_[:indice_]
        nb_              = np.array(nb_)
        x_              = np.array(x_)
        y_              = np.array(y_)
        t_              = np.array(t_)

        self.nb_         = nb_
        self.x_          = x_
        self.y_          = y_
        self.t_          = t_



        self.indice_final_ = indice_




############################################################
############################################################


def random_walk_2d(*args, **kwargs):
    """
    Generate 2D trajectories.

    Arguments:

        lambda (float): stochastic integral lambda = 0 ito, 1/2 stratonovich, 1 klimintovitch(?)
        sigma (float):  width of the trap
        sigma_noise (float):   positioning noise
        dt (float):     time between frames
        n_short (int):  number of small time steps between the returned time steps
        N_mean (int):   mean duration of the trajectories in number of points
        n_trajs (int):  number of individual trajectories
        L (float):      size of the trap in microns (half)
        density (float):       effective density of the random walks
        D0 (float):     local diffusivity
        amplitude_D (float):   amplitude of diffusivity variation
        amplitude_V (float):   ampltiude of the trap or slope
        mode_D (str):   any of '*linear*', '*gaussian_trap*', '*gaussian_bump*'
        mode_V (str):   any of '*potential_force*', '*potential_linear*'
        mode_gamma (str):       any of '*equilibrium*', '*fixed*'
        verbose (bool): print trajectory index (default is False)

    Returns:

        pandas.DataFrame:
            column '*n*' is trajectory index, columns '*x*' and '*y*' are spatial
            coordinates (in micrometers) and column '*t*' is time (in seconds).

    See also :class:`generate_trajectories` for default values.

    """
    # rename the arguments with a trailing underscore
    for arg in list(kwargs.keys()):
        if arg in ('lambda', 'sigma', 'sigma_noise', 'dt', 'n_short', 'N_mean', 'n_trajs',
                'L', 'density', 'D0', 'amplitude_D', 'amplitude_V'):
            kwargs[arg+'_'] = kwargs.pop(arg)
    # turn verbosity off by default
    kwargs['verbose'] = kwargs.get('verbose', False)
    # generate trajectories
    gen = generate_trajectories(*args, **kwargs)
    gen.generate_the_actual_trajectories()
    gen.noisify()
    # return a DataFrame
    gen.create_traj()
    return gen.traj


############################################################
############################################################


if __name__ == '__main__':

    #“”””
    #One imagine that some of those words were attached to actual meaning of some sort.
    #“””

    import warnings
    warnings.filterwarnings('ignore')

    # check the parameters np and pandas for weird bug
    print(np.__version__)
    print(pd.__version__)

    # parameters to be changed to generate
    lambda_         = 1 # stochastic integral lambda = 0 ito, 1/2 strato , 1 klimintovitvh
    sigma_          = 0.4 # width of the trap
    sigma_noise_    = 0.010 # positioning noise
    dt_             = 25e-3 # time between frames
    dt_space_       = 10 * dt_ # time between trajectories
    n_short_        = 1000 # number of small time steps between the time stepds oututed
    N_mean_         = 5 # mean duration of the traj in number of poitns
    n_trajs_        = 200 # number of individual trajs
    L_              = 0.5 # size of the trap in microns (half)
    dt_short_       = dt_ /n_short_ # short time between steps in the generation trajs
    density_        = 2. # effective desnity of RW
    Nb_tot_loc_     = density_*L_*L_ #nombre tot
    N_duration_tot_ = np.round(n_trajs_ /Nb_tot_loc_) # estimate duration
    #############################################
    D0_             = 0.1 # local diffusivity
    amplitude_D_    = 0.02 # amplitude of diffusivity variation
    amplitude_V_    = 5 # ampltidue of the trap
    #############################################
    mode_D          ="gaussian_bump"
    mode_V          ="potential_force"
    mode_gamma      = "equilibrium"
    name_file_out   ="trajectories.txt"

    gen = generate_trajectories(lambda_ = lambda_,
        sigma_ = sigma_,
        sigma_noise_ = sigma_noise_,
        dt_ = dt_,
        #dt_space_ = dt_space_,
        n_short_ = n_short_,
        N_mean_ = N_mean_,
        n_trajs_ = n_trajs_,
        L_ = L_,
        #dt_short_ = dt_short_,
        density_ = density_,
        #Nb_tot_loc_ = Nb_tot_loc_,
        #N_duration_tot_ = N_duration_tot_,
        D0_ = D0_,
        amplitude_D_ = amplitude_D_,
        amplitude_V_ = amplitude_V_,
        mode_D = mode_D,
        mode_V = mode_V,
        mode_gamma = mode_gamma,
        name_file_out = name_file_out )
    gen.generate_the_actual_trajectories()
    gen.noisify()
    gen.create_traj()
    gen.print_traj()

    #print(gen.nb_.size)
    #print(gen.x_.size)
    #print(gen.y_.size)
    #print(gen.t_.size)




# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import numpy as np
import pandas as pd

from tramway.inference.base import Maps
from tramway.inference.bayes_factors.get_D_posterior import *
from tramway.inference.gradient import setup_with_grad_arguments, get_grad_kwargs
from tramway.inference.snr import add_snr_extensions


setup = {
    'name': 'd.conj_prior',
    'provides': 'snr',
    'infer': 'infer_d_conj_prior',
    'arguments': dict(
        localization_error=(
            '-e', dict(type=float, help='localization error (same units as the variance)')),
        alpha=dict(type=float, default=.95, help='confidence level')),
}
setup_with_grad_arguments(setup)

def infer_d_conj_prior(cells, alpha=.95, return_zeta_spurious=True, trust=False, **kwargs):
    """
    Infer diffusivity MAP and confidence interval using a conjugate prior [Serov et al. 2019]

    Arguments:

        cells (tramway.inference.base.Distributed): distributed cells.

        alpha (float): confidence level for CI estimation.

        return_zeta_spurious (bool): add variable *zeta_spurious* to the returned dataframe.

        trust (bool): if ``False``, silently catch any exception raised in `get_D_confidence_interval`.

    Returns:

        pandas.DataFrame: maps including diffusivity MAPs and confidence intervals, and SNR extensions.

    Other valid input arguments are localization- and gradient-related arguments.

    See also :func:`~tramway.inference.bayes_factors.get_D_posterior.get_D_confidence_interval`,
    :meth:`~tramway.inference.base.Local.get_localization_error`
    and :func:`~tramway.inference.gradient.get_grad_kwargs`.
    """
    dim = cells.dim
    dt = cells.any_cell().dt[0]
    sigma2 = cells.get_localization_error(kwargs)
    if sigma2 is None:
        raise ValueError('undefined localization precision; please define `sigma` or `sigma2`')
    maps = add_snr_extensions(cells, _zeta_spurious=False)
    n, zeta_t, V, V_pi = maps['n'], Maps(maps)['zeta_total'], maps['V'], maps['V_prior']
    index, D_map, D_ci = [], [], []
    for i in cells:
        try:
            n_i = n.loc[i]
            zeta_i = zeta_t.loc[i]
            V_i = V.loc[i]
            V_minus_i = V_pi.loc[i]
        except KeyError:
            continue
        try:
            _map, _ci = get_D_confidence_interval(
                alpha, n_i, zeta_i, V_i, V_minus_i, dt, sigma2, dim)
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            if trust:
                raise
            continue
        index.append(i)
        D_map.append(_map)
        D_ci.append(_ci[np.newaxis, :])
    D_map = np.array(D_map)
    D_ci = np.vstack(D_ci)
    if return_zeta_spurious:
        reverse_index = np.full(cells.adjacency.shape[0], -1, dtype=int)
        reverse_index[index] = np.arange(len(index))
        grad_kwargs = get_grad_kwargs(**kwargs)
        g_index, g = [], []
        sd = maps['sd']
        g_defined = np.zeros(len(sd), dtype=bool)
        for j, i in enumerate(sd.index):
            gradD = cells.grad(i, D_map, reverse_index, **grad_kwargs)
            if gradD is not None:
                # the nnz condition does not prevent the gradient to be defined
                # but such cases are excluded anyway in the calculation of zeta_spurious
                g_defined[j] = True
                g_index.append(i)
                g.append(gradD[np.newaxis, :])
        g = np.concatenate(g, axis=0)
        # zeta_spurious
        dt = cells[i].dt[0]  # `add_snr_extensions` checked that all dts are equal
        sd = sd.values[:, np.newaxis]
        zeta_spurious = g * dt / sd[g_defined]
        maps = maps.join(pd.DataFrame(
            zeta_spurious,
            index=g_index,
            columns=['zeta_spurious ' + col for col in cells.space_cols],
        ))
    maps = maps.drop(columns=['sd']).join(pd.DataFrame(
        np.hstack((D_map[:, np.newaxis], D_ci)),
        index=index,
        columns=['diffusivity', 'ci low', 'ci high']))
    return maps

def infer_d_map_and_ci(*args, **kwargs):
    """
    Backward-compatibility alias for :func:`infer_d_conj_prior`.
    """
    return infer_d_conj_prior(*args, **kwargs)

__all__ = ['infer_d_conj_prior', 'infer_d_map_and_ci', 'setup']



# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from ..artefact import analysis
from .abc import *
from tramway.tessellation import Partition


class BaseSampler(AnalyzerNode):
    """ wraps the `Partition.cell_index` logics that readily implements
    all the possibilities mentioned in this module.
    """
    __slots__ = ('_min_location_count',)
    def __init__(self, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        self._min_location_count = 0
    @property
    def min_location_count(self):
        """
        *int*: Minimum number of assigned (trans-)locations, as given
            by the default Voronoi sampling;
            any microdomain with too few assigned (trans-)locations
            are marked as disabled and are excluded from further analyses
        """
        return self._min_location_count
    @min_location_count.setter
    def min_location_count(self, n):
        self._min_location_count = n
    @analysis(2, 'segmentation')
    def sample(self, spt_dataframe, segmentation=None, **kwargs):
        df = spt_dataframe
        if self.min_location_count is not None:
            kwargs['min_location_count'] = self.min_location_count
        if segmentation is None:
            if self.tesseller.initialized:
                segmentation = self.tesseller.tessellate(spt_dataframe)
            if self.time.initialized:
                segmentation = self.time.segment(spt_dataframe, segmentation)
        if segmentation is None:
            cell_index = None
        else:
            cell_index = segmentation.cell_index(df, **kwargs)
        sample = Partition(df, segmentation, cell_index)
        try:
            self.tesseller.bc_update_params(sample.param)
        except AttributeError:
            pass
        if kwargs:
            sample.param['partition'] = kwargs
        return sample
    @property
    def tesseller(self):
        return self._parent.tesseller
    @property
    def time(self):
        return self._parent.time

Sampler.register(BaseSampler)

class VoronoiSampler(BaseSampler):
    """ nearest neighbor assignment.
    
    Each data point is assigned to exactly one cell/microdomain."""
    __slots__ = ()
    def sample(self, df, segmentation=None):
        parent = super(VoronoiSampler, self)
        return parent.sample(df, segmentation)

#Sampler.register(VoronoiSampler)


class SamplerInitializer(Initializer):
    """ Initializer class for the :class:`~tramway.analyzer.RWAnalyzer`
    :attr:`~tramway.analyzer.RWAnalyzer.sampler` main attribute.

    The :attr:`~tramway.analyzer.RWAnalyzer.sampler` attribute self-modifies
    on calling any of the *from_...* methods.

    The default not-initialized :attr:`~tramway.analyzer.RWAnalyzer.sampler`
    attribute behaves like:

    .. code-block:: python

        a = RWAnalyzer()
        a.sampler.from_voronoi()

    """
    __slots__ = ()
    def from_voronoi(self):
        """ default sampler.

        The data points are assigned to the nearest cell/microdomain.
        
        See also :class:`VoronoiSampler`."""
        self.specialize( VoronoiSampler )
    def from_spheres(self, radius):
        """ nearest neighbor assignment by distance.

        Voronoi assignment is applied first and then, for each microdomain,
        the maximum distance of the assigned points to the center of the cell is considered.
        The `radius` argument defines a lower and/or upper bound(s) on the distance
        from the cell center.
        If the most distant assigned point is out of the specified bounds,
        additional points are selected (or assigned points are discarded)
        in order of distance from the corresponding cell center.
        
        See also :class:`SphericalSampler`."""
        self.specialize( SphericalSampler, radius )
    def from_nearest_neighbors(self, knn):
        """ nearest neighbor assignment by count.

        The Voronoi assignment is applied first.
        The `knn` argument defines a lower and/or upper bound(s) on the number of neighbors.
        For each microdomain, if the number that results from Voronoi assignment is
        out of the specified bounds, additional points are selected (or assigned points are
        discarded) in order of distance from the corresponding cell center.
        
        See also :class:`Knn`."""
        self.specialize( Knn, knn )
    def from_nearest_time_neighbors(self, knn):
        """ similar to `from_nearest_neighbors` but cell/microdomain size is adjusted in time
        instead of space.
        
        See also :class:`TimeKnn`."""
        self.specialize( TimeKnn, knn )
    def sample(self, spt_data, segmentation=None):
        """ main processing method. """
        self.from_voronoi()
        return self._sampler.sample(spt_data, segmentation)
    @property
    def _sampler(self):
        return self._parent.sampler
    @property
    def min_location_count(self):
        raise AttributeError("'SamplerInitializer' object has no attribute 'min_location_count'")
    @min_location_count.setter
    def min_location_count(self, n):
        self.from_voronoi()
        self._sampler.min_location_count = n


class SphericalSampler(BaseSampler):
    """ nearest neighbor selection by distance."""
    __slots__ = ('_radius',)
    def __init__(self, radius, **kwargs):
        BaseSampler.__init__(self, **kwargs)
        self._radius = radius
    @property
    def radius(self):
        """
        *float* or *(float, float)* or *callable*:
            Lower bound or (lower bound, upper bound) pair on the distance
            from a cell/microdomain center;
            alternatively, a function can define these bounds for each
            microdomain (takes the microdomain index as input argument)
        """
        return self._radius
    @radius.setter
    def radius(self, r):
        self._radius = r
    def sample(self, spt_data, segmentation=None):
        parent = super(SphericalSampler, self)
        return parent.sample(spt_data, segmentation, radius=self.radius)


class Knn(BaseSampler):
    """ nearest neighbor selection by count."""
    __slots__ = ('_knn',)
    def __init__(self, knn, **kwargs):
        BaseSampler.__init__(self, **kwargs)
        self._knn = knn
    @property
    def knn(self):
        """ 
        *int* or *(int, int)* or *callable*:
            Lower bound or (lower bound, upper bound) pair on the number of data points
            which, if Voronoi assignment yields a number out of these bounds,
            triggers a procedure that picks extra points (or discards assigned points)
            wrt distance from the Voronoi cell center.
            This applies independently for each microdomain with insufficient or too
            many points.
            Alternatively, bounds can be defined on a per-microdomain basis
            with a function that takes the microdomain index as an `int` and
            returns either an `int` or a pair of `int` or :const:`None`
        """
        return self._knn
    @knn.setter
    def knn(self, r):
        self._knn = r
    def sample(self, spt_data, segmentation=None):
        parent = super(Knn, self)
        return parent.sample(spt_data, segmentation, knn=self.knn,
                fast_min_nn=False)


class TimeKnn(BaseSampler):
    """ nearest neighbor selection by count across time."""
    __slots__ = ('_knn',)
    def __init__(self, knn, **kwargs):
        BaseSampler.__init__(self, **kwargs)
        self._knn = knn
    @property
    def knn(self):
        """
        *int* or *(int, int)* or *callable*:
            Lower bound or (lower bound, upper bound) pair on the number of data points
            which, if Voronoi assignment yields a number out of these bounds,
            triggers the shrinking or widening of the time window,
            centered on the original middle point of the time segment.
            This applies independently for each microdomain with insufficient or too
            many points.
            microdomain with insufficient or too many points.
            Alternatively, bounds can be defined on a per-microdomain basis
            with a function that takes the microdomain index as an `int` and
            returns either an `int` or a pair of `int` or :const:`None`
        """
        return self._knn
    @knn.setter
    def knn(self, r):
        self._knn = r
    def sample(self, spt_data, segmentation=None):
        parent = super(TimeKnn, self)
        return parent.sample(spt_data, segmentation, time_knn=self.knn)


__all__ = ['Sampler', 'SamplerInitializer', 'BaseSampler', 'VoronoiSampler',
        'SphericalSampler', 'Knn', 'TimeKnn']




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from .abc import *
from .proxy import TessellerProxy
from .plugin import TessellerPlugin
from . import proxied as tessellers
from .post import cell_mergers


class TessellerInitializer(Initializer):
    """
    Initializer class for the :class:`~tramway.analyzer.RWAnalyzer`
    :attr:`~tramway.analyzer.RWAnalyzer.tesseller` main attribute.

    The :attr:`~tramway.analyzer.RWAnalyzer.tesseller` attribute
    self-modifies on calling any of the *from_...* methods.

    The easiest way to define a tesseller is to set the
    :attr:`~tramway.analyzer.RWAnalyzer.tesseller` attribute with
    any of the tesseller classes provided by the :mod:`tessellers` module:

    .. code-block:: python

        from tramway.analyzer import *

        a = RWAnalyzer()
        a.tesseller = tessellers.KMeans

    Note that :mod:`tessellers` is made available by importing :mod:`tramway.analyzer`.

    """
    __slots__ = ()
    def from_plugin(self, plugin):
        self.specialize( TessellerPlugin, plugin )
    def from_callable(self, cls):
        """
        Argument:

            cls (callable):
                no-argument callable object (usually a class)
                that returns a :class:`Tesseller` object.

        """
        if isinstance(cls, type) and issubclass(cls, TessellerProxy):
            self.specialize( cls )
        else:
            self.specialize( TessellerProxy, cls )

    @property
    def _mpl_impl(self):
        from .mpl import Mpl
        return Mpl
    @property
    def mpl(self):
        """ tramway.analyzer.tesseller.mpl.Mpl: Matplotlib utilities """
        return self._mpl_impl(self)


__all__ = ['Tesseller', 'TessellerInitializer', 'TessellerProxy', 'TessellerPlugin', 'tessellers',
        'TessellationPostProcessing', 'cell_mergers']

